from __future__ import (
    annotations,
)  # python3.7+ feature to allow self-referencing type hints

from typing import Callable, Dict, List, Literal, Optional, Union
import collections
import dataclasses
import datetime
import getpass
import math
import os
import shutil
import socket
import subprocess

import pandas as pd

from realhf.base.constants import LOG_ROOT
from realhf.scheduler.client import JobException, JobInfo, JobState
import realhf.base.cluster
import realhf.base.logging as logging

logger = logging.getLogger("scheduler.slurm.utils")

SQUEUE_FIELDS = [
    "JobID",
    "State",
    "SubmitTime",
    "StartTime",
    "Name",
    "NodeList",
    "UserName",
    "MaxCPUs",
    "cpus-per-task",
    "NumTasks",
    "tres-alloc",
]
STATUS_MAPPING = {
    "RUNNING": JobState.RUNNING,
    "COMPLETING": JobState.RUNNING,
    "PENDING": JobState.PENDING,
    "CANCELLED": JobState.CANCELLED,
    "FAILED": JobState.FAILED,
    "COMPLETED": JobState.COMPLETED,
    "OUT_OF_MEMORY": JobState.FAILED,
    "DEADLINE": JobState.COMPLETED,
    "TIMEOUT": JobState.COMPLETED,
}


class SlurmResourceNotEnoughException(Exception):
    pass


# dataclasses


@dataclasses.dataclass
class SlurmResource:
    # a data class that represents a slurm resource quota
    mem: int = 0
    cpu: int = 0
    gpu_type: Optional[Literal["tesla", "geforce"]] = None
    gpu: Union[float, int] = 0

    def __check_gpu_type(self, other: SlurmResource) -> str:
        self_gpu_type = None if self.gpu == 0 else self.gpu_type
        other_gpu_type = None if other.gpu == 0 else other.gpu_type
        valid_gpu_type = self_gpu_type == other_gpu_type or (
            self_gpu_type or other_gpu_type
        )
        assert (
            valid_gpu_type
        ), f"Cannot add two different gpu types {self_gpu_type}, {other_gpu_type}."
        return self_gpu_type if self_gpu_type else other_gpu_type

    def __str__(self):
        return (
            "SlurmResource: \n"
            + "mem: "
            + str(self.mem)
            + " MB \n"
            + "cpu: "
            + str(self.cpu)
            + " \n"
            + "gpu: "
            + str(self.gpu)
            + " \n"
            + "gpu_type: "
            + str(self.gpu_type)
        )

    def __mul__(self, other: int) -> SlurmResource:
        assert isinstance(
            other, int
        ), "ResourceRequirement can only be multiplied by int."
        return SlurmResource(
            mem=self.mem * other,
            cpu=self.cpu * other,
            gpu=self.gpu * other,
            gpu_type=self.gpu_type,
        )

    def __rmul__(self, other: int) -> SlurmResource:
        return self.__mul__(other)

    def __add__(self, other: SlurmResource) -> SlurmResource:
        assert isinstance(
            other, SlurmResource
        ), "SlurmResource can only add another SlurmResource instance."
        return SlurmResource(
            mem=self.mem + other.mem,
            cpu=self.cpu + other.cpu,
            gpu=self.gpu + other.gpu,
            gpu_type=self.__check_gpu_type(other),
        )

    def __sub__(self, other: SlurmResource) -> SlurmResource:
        assert isinstance(
            other, SlurmResource
        ), "SlurmResource can only subtract another SlurmResource instance."
        return SlurmResource(
            mem=self.mem - other.mem,
            cpu=self.cpu - other.cpu,
            gpu=self.gpu - other.gpu,
            gpu_type=self.__check_gpu_type(other),
        )

    def __neg__(self) -> SlurmResource:
        return SlurmResource(
            mem=-self.mem, cpu=-self.cpu, gpu=-self.gpu, gpu_type=self.gpu_type
        )

    def __eq__(self, other: SlurmResource) -> bool:
        return (
            self.mem == other.mem
            and self.cpu == other.cpu
            and self.gpu == other.gpu
            and self.gpu_type == other.gpu_type
        )

    def __lt__(self, other: SlurmResource) -> bool:
        if self.gpu_type != other.gpu_type:
            if self.gpu_type is None:
                return True
            if self.gpu_type == "geforce":
                return other.gpu_type == "tesla"
            if self.gpu_type == "tesla":
                return False
        if self.gpu != other.gpu:
            return self.gpu < other.gpu
        if self.cpu != other.cpu:
            return self.cpu < other.cpu
        if self.mem != other.mem:
            return self.mem < other.mem

    def valid(self) -> bool:
        # check if it is a valid resource requirement
        if self.gpu_type not in ["geforce", "tesla", None]:
            return False
        if self.mem < 0 or self.cpu < 0 or self.gpu < 0:
            return False
        return True


@dataclasses.dataclass
class SlurmLaunchInfo:
    """A SlurmLaunchInfo contains all informantion required to **launch** a slurm job.

    Matching one `TasksGroup` in `SchedulingConfig` and one slurm job.

    The naming conventions:
        - `job`: Literally a slurm job with a (maybe non-unique) job name and an unique job ID,
            which may contain multiple job steps and processes. It corresponds an `sbatch` or `srun` call.
            Job names are guaranteed to be unique using the scheduler within this repo.
        - `jobstep`: Literally a slurm job step with a unique job step ID, i.e., ${jobID}.${stepID},
            which corresponds to a running instance `apps.remote` script, but may still contain multiple processes.
            A job step occupies at most one GPU. Processes in the same job step must share the same GPU.
        - `wproc`: A single worker process launched by `apps.remote` script, which may occupy less than 1 GPU.
            A worker just corresponds to a process.
        - `task`: The alias of `jobstep`. It is easier to under stand this concept in the context of `srun` command.
            `--ntasks' is just the number of jobsteps. We use the alternative term `jobstep` to avoid confusion.

    Attributes:
        run_name (str): Identifier of this run, typically ${exp_name}_${trial_name}.
        worker_type (str): Type of workers to be launched, e.g. model_worker, data_worker, etc.
        worker_submission_idx (int): For heterogeneous scheduling, we submit jobs of the same worker_type to slurm
            for multiple times. `worker_submission_idx` is used to distinguish them, so the (global) slurm job name will
            be ${run_name}:${worker_type}:${worker_submission_idx}.
        wprocs_in_job: The number of worker processes in this slurm job (of all job steps).
        n_jobsteps (int): The number of job steps of this slurm job. This is also the group size of the multiprog file.
            Will be resolved automatically according to GPU requirement.
        wprocs_per_jobstep: The number of worker processes in each job step, as well as the number of sub-processes
            spawned by `apps.remote`. Will be resolved automatically according to GPU requirement.

        resource_requirement (SlurmResource): The resource requirement of this job, including all job steps.
        cmd (str): The command to be executed.
        container_image (str): .
        container_mounts (str): .
        env_vars (dict): .
        node_type (str): .
        nodelist (str): .
        exclude (str): .
        partition (str, optional): .
        time_limit (str, optional): Slurm job time limit.
        begin (str, optional): Scheduled worker start time.
        deadline (str, optional): Scheduled worker end time.
        hostfile (bool): Whether to use hostfile for `--distribution=arbitrary` scheduling.
        hostfile_content (str, optional): The content of the hostfile.
        multiprog (bool): Whether to use multiprog file for `--multi-prog` job submission.
        multiprog_content (str, optional): The content of the multiprog file.
    """

    run_name: str
    exper_name: str
    trial_name: str
    worker_type: str
    worker_submission_idx: int
    wprocs_in_job: int

    resource_requirement: SlurmResource
    cmd: str
    container_image: str
    container_mounts: str
    env_vars: dict
    node_type: str
    nodelist: str
    exclude: str
    partition: Optional[str] = None
    time_limit: Optional[str] = None
    begin: Optional[str] = None
    deadline: Optional[str] = None
    # hostfile
    hostfile: bool = True
    hostfile_content: Optional[str] = None
    # multiprog options, override cmd
    multiprog: bool = True
    multiprog_content: Optional[str] = None

    n_jobsteps: int = None
    wprocs_per_jobstep: int = None

    job_info: Optional[JobInfo] = None

    def __post_init__(self):
        """resolve fractional GPU resource requirement"""
        gpu_per_worker = self.resource_requirement.gpu
        # assert gpu_per_worker <= 1 and gpu_per_worker >= 0
        if gpu_per_worker < 1 and gpu_per_worker > 0:
            self.resource_requirement.gpu = 1
            self.wprocs_per_jobstep = math.floor(1 / gpu_per_worker)
            self.resource_requirement.cpu *= self.wprocs_per_jobstep
            self.resource_requirement.mem *= self.wprocs_per_jobstep
            self.n_jobsteps = math.ceil(
                self.wprocs_in_job / self.wprocs_per_jobstep
            )
            logger.info(
                f"Resolved fractional GPU requirement for {self.slurm_name}"
            )
            logger.info(
                f"GPU per worker {gpu_per_worker}, workers per jobstep (process size in `apps.remote`) {self.wprocs_per_jobstep}, "
                f"number of jobsteps (instance of running `apps.remote`) {self.n_jobsteps}"
            )
        elif gpu_per_worker == 0:
            self.wprocs_per_jobstep = self.wprocs_in_job
            self.n_jobsteps = 1
        elif gpu_per_worker == 1:
            self.n_jobsteps = self.wprocs_in_job
            self.wprocs_per_jobstep = 1
        else:
            self.n_jobsteps = 1
            self.wprocs_per_jobstep = 1

    @property
    def slurm_name(self) -> str:
        return (
            f"{self.run_name}:{self.worker_type}:{self.worker_submission_idx}"
        )

    @property
    def slurm_id(self) -> Optional[str]:
        if self.job_info:
            return self.job_info.slurm_id
        else:
            return None

    @property
    def log_path(self) -> str:
        return os.path.join(
            LOG_ROOT,
            self.exper_name,
            self.trial_name,
            f"{self.worker_type}-{self.worker_submission_idx}",
        )

    @property
    def multiprog_path(self) -> str:
        return os.path.join(
            LOG_ROOT,
            self.exper_name,
            self.trial_name,
            f"{self.worker_type}-{self.worker_submission_idx}.multiprog",
        )

    @property
    def hostfile_path(self) -> str:
        return os.path.join(
            LOG_ROOT,
            self.exper_name,
            self.trial_name,
            f"{self.worker_type}-{self.worker_submission_idx}.hostfile",
        )

    def show_log(self):
        try:
            terminal_columns = os.get_terminal_size().columns
        except OSError:
            terminal_columns = shutil.get_terminal_size().columns
        logger.info(
            f"Showing log of slurm job: {self.worker_type}-{self.worker_submission_idx}\n\n{'-'*terminal_columns}"
        )
        subprocess.Popen(["tail", "-n50", self.log_path]).wait(timeout=3)
        logger.info(
            f"End of log: {self.worker_type}-{self.worker_submission_idx}\n\n{'-'*terminal_columns}"
        )

    def update(self):
        job_infos = query_jobs(slurm_names=[self.slurm_name])
        job_infos = sorted(
            job_infos,
            key=lambda x: parse_formatted_time(x.submit_time),
            reverse=True,
        )
        self.job_info = job_infos[0] if len(job_infos) > 0 else None
        if self.job_info:
            return self.job_info.state
        else:
            return None

    def cancel(self, signal: Literal["SIGINT", "SIGKILL"] = "SIGKILL"):
        cancel_jobs(slurm_names=[self.slurm_name], signal=signal)
        self.job_info = JobInfo(name=self.slurm_name, state=JobState.CANCELLED)

    def __str__(self):
        s = f"SlurmLaunchInfo [{self.slurm_name}] \n"
        s += f"Resources: [\n{self.resource_requirement}\n]\n"
        s += f"Multiprog Filepath: [{self.multiprog_path}]\n"
        s += f"Multiprog Content: [\n{self.multiprog_content}\n]\n"
        s += f"Hostfile Filepath: [{self.hostfile_path}]\n"
        s += f"Hostfile Content: [\n{self.hostfile_content}\n]\n"
        if self.job_info is None:
            job_info_str = "None"
        else:
            job_info_str = "\n".join(
                [f"{k}: {v}" for k, v in self.job_info.__dict__.items()]
            )
        s += f"Runtime JobInfo: [\n{job_info_str}\n]\n"
        env_var_str = "\n".join([f"{k}: {v}" for k, v in self.env_vars.items()])
        s += f"Env vars: [\n{env_var_str}\n]\n"
        return s

    def commit(self):
        os.makedirs(os.path.dirname(self.log_path), exist_ok=True, mode=0o775)

        ntasks = self.n_jobsteps
        mem = self.resource_requirement.mem
        cpu = self.resource_requirement.cpu
        gpu = self.resource_requirement.gpu

        cmd = self.cmd

        # assert gpu == 1 or gpu == 0, "Slurm job GPU requirement should be resolved to a integer."
        gpu_type = self.resource_requirement.gpu_type

        if self.multiprog:
            with open(self.multiprog_path, "w") as f:
                f.write(self.multiprog_content)
        if self.hostfile:
            with open(self.hostfile_path, "w") as f:
                f.write(self.hostfile_content)

        logger.info(
            f'Allocating {ntasks} jobstep(s) "{self.worker_type}" submission index {self.worker_submission_idx}'
            f" with {cpu} cpu, {gpu} gpu and {mem} MB memory."
        )
        logger.info(f"To check the output, run \n\t`tail -f {self.log_path}`.")

        # Setup sbatch
        # head
        lines = [
            "#!/bin/bash",
            f"#SBATCH --job-name={self.slurm_name}",
            f"#SBATCH --output={self.log_path}",
            f"#SBATCH --ntasks={ntasks}",
            f"#SBATCH --gpus-per-task={gpu_type}:{gpu}" if gpu >= 1 else "",
            f"#SBATCH --cpus-per-task={cpu}",
            f"#SBATCH --mem-per-cpu={mem // max(1, cpu)}M",
            f"#SBATCH --partition={self.partition}" if self.partition else "",
            "#SBATCH --distribution=arbitrary" if self.hostfile else "",
            # f'#SBATCH --nodelist={spec.nodelist}' if spec.nodelist is not None else "",
            # f'#SBATCH --exclude={spec.exclude}' if spec.exclude is not None else "",
            f"#SBATCH --time={self.time_limit}" if self.time_limit else "",
            f"#SBATCH --begin={self.begin}" if self.begin else "",
            f"#SBATCH --deadline={self.deadline}" if self.deadline else "",
        ]

        srun_env = os.environ.copy()
        if self.hostfile:
            srun_env["SLURM_HOSTFILE"] = self.hostfile_path
        # Setup step command.
        # add current directory into container mounts to ensure editable mode for realhf package
        container_mounts = (
            f"{os.environ.get('REAL_PACKAGE_PATH', '$PWD')}:/realhf,"
            + self.container_mounts
        )
        srun_flags = [
            f"--ntasks={ntasks}",
            f"--cpus-per-task={cpu}",
            f"--gpus-per-task={gpu_type}:{gpu}" if gpu >= 1 else "",
            f"--mem-per-cpu={mem // max(1, cpu)}",
            (
                f"--export={','.join(str(k)+'='+str(v) for k, v in self.env_vars.items())}"
                if self.env_vars
                else ""
            ),
            f"--multi-prog" if self.multiprog else "",
            f"--container-image={self.container_image}",
            f"--container-mounts={container_mounts}",
            f"--container-mount-home",
        ]

        if self.multiprog:
            srun_cmd = f'srun -l {" ".join(srun_flags)} {self.multiprog_path}'
        else:
            srun_cmd = f'srun -l {" ".join(srun_flags)} {cmd}'

        lines += [
            'echo "[Runner] StartTime: $(date -u)"',
            'echo "[Runner] Host: $(hostname)"',
            "echo '[Runner] Command: {}'".format(srun_cmd),
            "echo '[Runner] Log: {}'".format(self.log_path),
            'echo "[Runner] CudaVisible: $CUDA_VISIBLE_DEVICES"',
            'echo "[Runner] CudaMpsPerc: $CUDA_MPS_ACTIVE_THREAD_PERCENTAGE"',
            srun_cmd,
            "RETCODE=$?",
            'echo "[Runner] FinishTime: $(date -u)"',
            'echo "[Runner] RetCode: $RETCODE"',
            'echo "[Runner] ------------"',
            "exit $RETCODE",
        ]

        script_strs = "\n".join(list(filter(lambda x: x, lines))) + "\n"
        script = script_strs.encode("ascii")

        def pad_output_str_to_length(s: str, pad_s: str, length: int):
            assert len(pad_s) == 1
            assert len(s) + 2 <= length
            n_pads = (length - len(s) - 2) // 2
            return pad_s * n_pads + " " + s + " " + pad_s * n_pads

        with open(self.log_path, "a") as f:
            f.write(
                pad_output_str_to_length("SBATCH SCRIPT BEGIN", "=", 80) + "\n"
            )
            f.write(script_strs)
            f.write(
                pad_output_str_to_length("SBATCH SCRIPT END", "=", 80) + "\n"
            )
            f.write(
                pad_output_str_to_length("SBATCH JOB INFO BEGIN", "=", 80)
                + "\n"
            )
            f.write(str(self))
            f.write(
                pad_output_str_to_length("SBATCH JOB INFO END", "=", 80) + "\n"
            )
            f.write(
                pad_output_str_to_length("JOB OUTPUT BEGIN", "=", 80) + "\n"
            )
        r = (
            subprocess.check_output(
                ["sbatch", "--parsable"], input=script, env=srun_env
            )
            .decode("ascii")
            .strip()
        )
        self.job_info = JobInfo(name=self.slurm_name, state=JobState.PENDING)


def parse_formatted_time(time_string: str) -> int:
    if time_string == "N/A":
        return -1
    d = datetime.datetime.strptime(time_string, "%Y-%m-%dT%H:%M:%S")
    return int(datetime.datetime.timestamp(d))


def unparse_formatted_time(timestamp: int) -> str:
    if timestamp == -1:
        return "N/A"
    d = datetime.datetime.fromtimestamp(timestamp)
    return d.strftime("%Y-%m-%dT%H:%M:%S")


# slurm command execute and output parsing
def query_jobs(
    slurm_names: Optional[List[str]] = None,
    slurm_ids: Optional[List[str]] = None,
    status: str = "all",
    delimiter: str = "__PSI__",
) -> List[JobInfo]:
    squeue_format = f":.{delimiter},".join(SQUEUE_FIELDS)
    cmd = ["squeue", "-O", squeue_format, f"-t{status}"]
    if slurm_names is not None:
        cmd += ["-n", ",".join(slurm_names)]
    if slurm_ids is not None:
        cmd += ["-j", ",".join([str(s) for s in slurm_ids])]
    output = (
        subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
        .decode("ascii")
        .strip()
    )
    rs = []
    for line in output.split("\n")[1:]:
        job_id, state, submit_time, start_time, slurm_name, nodelist, *_ = (
            line.split(delimiter)
        )
        rs.append(
            JobInfo(
                name=slurm_name,
                state=STATUS_MAPPING[state],
                host=nodelist,
                submit_time=submit_time,
                start_time=start_time,
                slurm_id=job_id.strip(),
            )
        )
    return rs


def cancel_jobs(
    slurm_names: Optional[List[str]] = None,
    slurm_ids: Optional[List[str]] = None,
    signal: Literal["SIGINT", "SIGKILL"] = "SIGKILL",
):
    assert (
        slurm_names is not None or slurm_ids is not None
    ), "Must specify slurm_names or slurm_ids."
    assert not (
        slurm_names and slurm_ids
    ), "Cannot specify both slurm_names and slurm_ids."
    cmd = ["scancel", "-s", signal]
    if slurm_names is not None:
        cmd += ["-n", ",".join(slurm_names)]
    elif slurm_ids is not None:
        cmd += ["-j", ",".join([str(s) for s in slurm_ids])]
    subprocess.check_call(cmd)
    logger.info(
        f"Cancelled Slurm job with signal {signal}: "
        f"slurm identifiers {slurm_names if slurm_ids is None else slurm_ids}"
    )


def parse_output_status_line(status):
    assert status.startswith("State=")
    status = status.split(" ")[0]
    status = status.split("=")[1]
    return status.split("+")


def parse_output_tres_line(tres):
    tres = tres.split("=", maxsplit=1)[1]
    tres = tres.split(",")
    res = SlurmResource()
    if len(tres) == 0:
        return SlurmResource()
    for t in tres:
        if t.startswith("mem"):
            if t.endswith("M"):
                res.mem = float(t.split("=")[1].strip("M"))
            elif t.endswith("G"):
                res.mem = float(t.split("=")[1].strip("G")) * 1024
            else:
                raise ValueError("Unknown memory unit.")
        elif t.startswith("cpu"):
            res.cpu = int(t.split("=")[1])
        elif t.startswith("gres/gpu"):
            prefix, sgpu = t.split("=")
            if ":" in prefix:
                res.gpu_type = prefix.split(":")[1]
            else:
                res.gpu = int(sgpu)
    return res


def _parse_output_status_line(status):
    assert status.startswith("State=")
    status = status.split(" ")[0]
    status = status.split("=")[1]
    return status.split("+")


def _parse_output_tres_line(tres):
    tres = tres.split("=", maxsplit=1)[1]
    tres = tres.split(",")
    res = SlurmResource()
    if len(tres) == 0:
        return SlurmResource()
    for t in tres:
        if t.startswith("mem"):
            if t.endswith("M"):
                res.mem = int(t.split("=")[1].strip("M"))
            elif t.endswith("G"):
                res.mem = int(float(t.split("=")[1].strip("G")) * 1024)
            else:
                raise ValueError("Unknown memory unit.")
        elif t.startswith("cpu"):
            res.cpu = int(t.split("=")[1])
        elif t.startswith("gres/gpu"):
            prefix, sgpu = t.split("=")
            if ":" in prefix:
                res.gpu_type = prefix.split(":")[1]
            else:
                res.gpu = int(sgpu)
    return res


def available_hostnames(
    node_type: Optional[List[str]] = None,
    nodelist: Optional[str] = None,
    exclude: Optional[str] = None,
) -> List[str]:
    all_nodelist: str = (
        subprocess.check_output('sinfo -o "%N" --noheader', shell=True)
        .decode("utf-8")
        .strip()
    )
    all_hostnames: List[str] = (
        subprocess.check_output(
            [
                "scontrol",
                "show",
                "hostnames",
                all_nodelist,
            ]
        )
        .decode("utf-8")
        .strip()
        .split("\n")
    )

    if nodelist is not None:
        valid_hostnames: List[str] = (
            subprocess.check_output(
                [
                    "scontrol",
                    "show",
                    "hostnames",
                    nodelist,
                ]
            )
            .decode("utf-8")
            .strip()
            .split("\n")
        )
    else:
        valid_hostnames = all_hostnames

    if exclude is not None:
        excluded_hostnames: List[str] = (
            subprocess.check_output(
                [
                    "scontrol",
                    "show",
                    "hostnames",
                    exclude,
                ]
            )
            .decode("utf-8")
            .strip()
            .split("\n")
        )
        for hn in excluded_hostnames:
            if hn in valid_hostnames:
                valid_hostnames.remove(hn)

    for hn in valid_hostnames:
        if hn not in all_hostnames:
            raise ValueError(
                f"Invalid host name: {hn}. Available host names: {all_hostnames}."
            )

    return list(
        filter(
            lambda x: realhf.base.cluster.node_name_is_node_type(x, node_type),
            valid_hostnames,
        )
    )


def get_all_node_resources() -> Dict[str, SlurmResource]:
    """Execute `scontrol show node` to get all node resources
    available in the slurm cluster.
    Return a list of SlurmResource
    """
    o = subprocess.check_output(["scontrol", "show", "node"]).decode("utf-8")
    nodes = o.split("\n\n")
    all_rres = {}
    for node in nodes:
        if len(node) <= 1:
            continue
        ls = node.split("\n")
        node_name = ls[0].split(" ")[0].split("=")[1]
        ctres = SlurmResource()
        atres = SlurmResource()
        for l in ls:
            l = l.strip("\n").strip()
            if l.startswith("State"):
                status = _parse_output_status_line(l)
                if (
                    "DOWN" in status
                    or "DRAIN" in status
                    or "NOT_RESPONDING" in status
                ):
                    break
            if l.startswith("CfgTRES"):
                ctres = _parse_output_tres_line(l)
            if l.startswith("AllocTRES"):
                atres = _parse_output_tres_line(l)
        ctres.gpu_type = atres.gpu_type = (
            realhf.base.cluster.spec.gpu_type_from_node_name(node_name)
        )
        rres = ctres - atres
        if rres.valid():
            all_rres[node_name] = rres
        else:
            all_rres[node_name] = SlurmResource(gpu_type=ctres.gpu_type)

    return all_rres


def resource_to_string(resources: Dict[str, SlurmResource]) -> str:
    resource_list = [
        {
            **{"NodeName": k},
            **{
                field.name: getattr(r, field.name)
                for field in r.__dataclass_fields__.values()
            },
        }
        for k, r in resources.items()
    ]
    return pd.DataFrame(resource_list).to_string(index=False)


def allocate_resources(
    infos: List[SlurmLaunchInfo],
    # strategy: Literal["pack", "plane"] = "pack",
) -> List[SlurmLaunchInfo]:
    """Allocate all slurm task specs, fill in the hostfile field of the specs
    Only support simple greedy strategy now, which allocates tasks to node with the most
    available resources without considering other tasks.
    """
    all_resources = get_all_node_resources()
    # all_resources = sorted([(k, v) for k, v in all_resources.items()],
    #                 key=lambda x: x[1],
    #                 reverse=True)
    infos = sorted(infos, key=lambda x: x.resource_requirement, reverse=True)
    for info_idx, info in enumerate(infos):
        valid_hostnames = available_hostnames(
            node_type=info.node_type,
            nodelist=info.nodelist,
            exclude=info.exclude,
        )
        valid_hostnames = list(
            filter(lambda x: x in all_resources, valid_hostnames)
        )
        valid_resources = {hn: all_resources[hn] for hn in valid_hostnames}
        valid_resources = sorted(
            valid_resources.items(), key=lambda x: x[1], reverse=True
        )
        task_left = info.n_jobsteps
        allocated = dict()
        for hostname, resource in valid_resources:
            tmp = task_left
            while task_left > 0:
                try:
                    resource = resource - info.resource_requirement
                except ValueError:
                    break
                if not resource.valid():
                    break
                task_left -= 1
            if tmp - task_left > 0:
                allocated[hostname] = tmp - task_left
            all_resources[hostname] = resource
        if task_left > 0:
            # logger.warning(f"Current resources in the cluster:\n {resource_to_string(get_all_node_resources())}")
            logger.warning(
                f'Unable to allocate {info.n_jobsteps} Jobs with name "{info.slurm_name}". '
                f"Resource Requirement of this job is: {dataclasses.asdict(info.resource_requirement)}. "
                f"Valid resources for this job is "
                f"(according to NodeType={info.node_type}, NodeList={info.nodelist}, "
                f"and Exclude={info.exclude}):\n {resource_to_string({k: v for k, v in get_all_node_resources().items() if k in valid_hostnames})}"
            )
            for pinfo in infos[:info_idx]:
                if (
                    len(
                        set(pinfo.hostfile_content.split("\n")).intersection(
                            set(valid_hostnames)
                        )
                    )
                    == 0
                ):
                    continue
                palloc = collections.defaultdict(lambda: 0)
                for _n in pinfo.hostfile_content.split("\n"):
                    palloc[_n] += 1
                logger.warning(
                    f'Found previous job "{pinfo.slurm_name}" (ntasks={pinfo.n_jobsteps}) '
                    f"has been allocated to the same set of nodes. "
                    f"Resource requirement of this job is: {dataclasses.asdict(pinfo.resource_requirement)}, "
                    f"allocation of this job is {dict(palloc)}."
                )
            raise SlurmResourceNotEnoughException()
        hostlist = []
        for hostname, task_num in allocated.items():
            hostlist += [hostname] * task_num
        info.hostfile_content = "\n".join(hostlist)
    return infos


def show_tesla():
    all_rres = get_all_node_resources()
    hostname = socket.gethostname()
    for k in available_hostnames(node_type=["a100"]):
        print(k, all_rres[k])


def show_all():
    all_rres = get_all_node_resources()
    for k, v in all_rres.items():
        print(k, v)


if __name__ == "__main__":
    show_tesla()



from collections import defaultdict
from typing import *
import dataclasses
import itertools
import os
import platform
import socket
import time

import realhf.base.logging as logging
import realhf.base.name_resolve as name_resolve
import realhf.base.names as names
import realhf.base.network as network

logger = logging.getLogger("System-GPU", "system")

GPU_DEVICES_ISOLATED = False
GLOBAL_PROCESS_GROUP_NAME = "master"


def gpu_count():
    """Returns the number of gpus on a node. Ad-hoc to frl cluster."""
    if platform.system() == "Darwin":
        return 0
    elif platform.system() == "Windows":
        try:
            import torch

            return torch.cuda.device_count()
        except ImportError:
            return 0
    else:
        dev_directories = list(os.listdir("/dev/"))
        for cnt in itertools.count():
            if "nvidia" + str(cnt) in dev_directories:
                continue
            else:
                break
        return cnt


def set_cuda_device(device):
    """Set the default cuda-device. Useful on multi-gpu nodes. Should be called in every gpu-thread."""
    # logger.info(f"Setting device to {device}.")
    if device != "cpu":
        import torch

        torch.cuda.set_device(device)


def reveal_ddp_identity(expr_name, trial_name, worker_index):
    master_group_name = names.trainer_ddp_peer(
        expr_name, trial_name, GLOBAL_PROCESS_GROUP_NAME
    )
    name_resolve.add_subentry(
        master_group_name, str(worker_index), keepalive_ttl=30
    )
    # local_peer_name = names.trainer_ddp_local_peer(expr_name, trial_name, socket.gethostname(), peer_name)
    # name_resolve.add_subentry(local_peer_name, peer_index, keepalive_ttl=30)


def isolate_cuda_device(
    worker_type: str,
    rank: int,
    world_size: int,
    experiment_name: str,
    trial_name: str,
):
    """Isolate CUDA_VISIBLE_DEVICES for each Slurm jobstep.

    To distinguish the concept of job/jobstep/worker/task, check scheduler/slurm/utils.py.
    A slurm job with multiple jobsteps will not set CUDA_VISIBLE_DEVICES properly.
    For example, if a job has 2 jobsteps, each with 1 GPU, and is allocated onto GPU 0 and 1,
    then CUDA_VISIBLE_DEVICES of these jobsteps will be 0,1, instead of 0 and 1.
    We use this function in `apps.remote` to isolate CUDA_VISIBLE_DEVICES for each jobstep.

    Note that this function is completely independent of `setup_ddp`.

    Args:
        worker_type (str): .
        rank (int): Rank of the **jobstep**.
        world_size (int): Size of the **jobsteps**, aka SLURM_NPROCS. However, we may call this function
            in other schedulers (e.g. local scheduler), so we don't use SLURM_NPROCS directly.
        experiment_name (str): .
        trial_name (str): .
    """
    if not os.environ.get("CUDA_VISIBLE_DEVICES"):
        return

    name_resolve_identifier = f"__type_{worker_type}"
    name_resolve.add_subentry(
        names.trainer_ddp_local_peer(
            experiment_name,
            trial_name,
            socket.gethostname(),
            name_resolve_identifier,
        ),
        rank,
        keepalive_ttl=60,
    )
    name_resolve.add_subentry(
        names.trainer_ddp_peer(
            experiment_name, trial_name, name_resolve_identifier
        ),
        rank,
        keepalive_ttl=30,
    )
    logger.debug(
        f"Worker type {worker_type} rank {rank} waiting for peers, world size {world_size}..."
    )
    while (
        len(
            name_resolve.get_subtree(
                names.trainer_ddp_peer(
                    experiment_name, trial_name, name_resolve_identifier
                )
            )
        )
        < world_size
    ):
        time.sleep(0.1)
    # logger.info(f"Rank {rank} discovers all peers, resolving local rank...")
    local_peer_name = names.trainer_ddp_local_peer(
        experiment_name,
        trial_name,
        socket.gethostname(),
        name_resolve_identifier,
    )
    local_peers = list(
        [
            str(x)
            for x in sorted(
                [int(x) for x in name_resolve.get_subtree(local_peer_name)]
            )
        ]
    )
    # logger.info(f"Rank {rank} discovers local peers with global ranks {local_peers}")

    local_peer_index = local_peers.index(str(rank))
    if len(os.environ["CUDA_VISIBLE_DEVICES"].split(",")) == len(local_peers):
        local_gpu_id = list(
            map(int, os.environ["CUDA_VISIBLE_DEVICES"].split(","))
        )[local_peer_index]
    elif len(os.environ["CUDA_VISIBLE_DEVICES"].split(",")) == 1:
        local_gpu_id = int(os.environ["CUDA_VISIBLE_DEVICES"])
    else:
        if not os.environ.get("REAL_MODE") == "LOCAL":
            raise RuntimeError(
                f"Unresolvable CUDA_VISIBLE_DEVICES {os.environ['CUDA_VISIBLE_DEVICES']} on host {network.gethostname()}, "
                f"local peers (global ranks) {local_peers}, local peer index {local_peer_index}."
            )
        devices = os.environ["CUDA_VISIBLE_DEVICES"].split(",")
        local_gpu_id = int(devices[local_peer_index % len(devices)])

    # logger.info(
    #     f"Worker type {worker_type} rank {rank} running on host {socket.gethostname()}, "
    #     f"local peer index: {local_peer_index}, local gpu id {local_gpu_id}."
    # )

    os.environ["CUDA_VISIBLE_DEVICES"] = str(local_gpu_id)
    os.environ["GPU_DEVICES_ISOLATED"] = "1"



from pathlib import Path
import importlib
import importlib.util
import os
import re
import sys

from .logging import getLogger

logger = getLogger("importing")


def import_module(path: str, pattern: re.Pattern):
    dirname = Path(path)
    for x in os.listdir(dirname.absolute()):
        if not pattern.match(x):
            continue
        module_path = os.path.splitext(os.path.join(dirname, x))[0]
        assert "realhf" in module_path
        start_idx = path.index("realhf")
        module_path = module_path[start_idx:]
        module_path = "realhf." + module_path.replace(os.sep, ".").replace(
            "realhf.", ""
        )
        logger.info(f"Automatically importing module {module_path}.")
        importlib.import_module(module_path)


def import_usercode(module_path: str, module_name: str):
    # Create a module spec
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    # Create a module object
    module = importlib.util.module_from_spec(spec)
    # Add the module to sys.modules
    sys.modules[module_name] = module
    # Execute the module in its own namespace
    spec.loader.exec_module(module)



# Request-reply stream between model workers and the master worker.
# The stream is composed of a pair of ZMQ sockets, one PUSH and one PULL, for asynchronous communication,
# i.e., the model worker can buffer requests from the master and execute them in any order under the hood.
from typing import Any, Dict, List, Optional, Tuple, Union
import asyncio
import dataclasses
import pickle
import re
import socket
import time
import uuid

import zmq

from realhf.base import logging, name_resolve, names
import realhf.api.core.system_api as system_api

logger = logging.getLogger("Request-Replay Stream")
ZMQ_IO_THREADS = 8

PUBSUB_BARRIER_NAME = "__pubsub_barrier__"


class NoMessage(Exception):
    pass


@dataclasses.dataclass
class Payload:
    handler: Union[system_api.ModelShardID, str]
    handle_name: str

    request_id: uuid.UUID = None
    syn_reply_id: uuid.UUID = None
    ack_reply_id: uuid.UUID = None

    send_time: float = None

    # Non-tensor data
    data: Any = None

    # RPC hooks
    pre_hooks: List[str] = dataclasses.field(default_factory=list)
    pre_hook_data: List[Any] = dataclasses.field(default_factory=list)

    post_hooks: List[str] = dataclasses.field(default_factory=list)
    post_hook_data: List[Any] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        if self.request_id is None:
            self.request_id = uuid.uuid4()
        if self.syn_reply_id is None:
            self.syn_reply_id = uuid.uuid4()
        if self.ack_reply_id is None:
            self.ack_reply_id = uuid.uuid4()


class NameResolvingRequstClient:

    def __init__(
        self,
        experiment_name: str,
        trial_name: str,
        n_subscribers: int,
        handler_routing: Dict[str | system_api.ModelShardID, int],
    ):

        self.context = zmq.Context.instance(io_threads=ZMQ_IO_THREADS)
        host_ip = socket.gethostbyname(socket.gethostname())

        self.send_sockets: List[zmq.Socket] = []
        for i in range(n_subscribers):
            s: zmq.Socket = self.context.socket(zmq.PUSH)
            send_port = s.bind_to_random_port(f"tcp://{host_ip}")
            s.setsockopt(zmq.LINGER, 0)

            master_send_name = names.request_reply_stream(
                experiment_name, trial_name, f"master_send_{i}"
            )
            name_resolve.add(
                name=master_send_name, value=f"{host_ip}:{send_port}"
            )
            logger.debug(
                f"Add master send address {host_ip}:{send_port} as {master_send_name}"
            )
            self.send_sockets.append(s)

        self.recv_socket: zmq.Socket = self.context.socket(zmq.PULL)
        recv_port = self.recv_socket.bind_to_random_port(f"tcp://{host_ip}")
        self.recv_socket.setsockopt(zmq.LINGER, 0)
        self.recv_address = f"{host_ip}:{recv_port}"

        master_recv_name = names.request_reply_stream(
            experiment_name, trial_name, "master_recv"
        )
        name_resolve.add(name=master_recv_name, value=self.recv_address)
        logger.debug(
            f"Add master send address {self.recv_address} as {master_recv_name}"
        )

        self._response_buffer: Dict[uuid.UUID, Payload] = {}
        self._handler_routing = handler_routing

        # master needs to wait all peers (subscribers) to connect
        while (
            len(
                name_resolve.get_subtree(
                    names.request_reply_stream(
                        experiment_name, trial_name, PUBSUB_BARRIER_NAME
                    )
                )
            )
            < n_subscribers
        ):
            time.sleep(0.1)
        logger.debug(
            f"Master discovered all {n_subscribers} "
            f"subscribers: {name_resolve.get_subtree(names.request_reply_stream(experiment_name, trial_name, PUBSUB_BARRIER_NAME))}."
        )

    def close(self):
        self.recv_socket.close()
        for send_socket in self.send_sockets:
            send_socket.close()
        self.context.destroy()

    def __del__(self):
        self.close()

    def post(self, payload: Payload) -> uuid.UUID:
        assert (
            payload.request_id is not None and payload.handle_name is not None
        )
        payload.send_time = time.monotonic()
        idx = self._handler_routing[payload.handler]
        self.send_sockets[idx].send(pickle.dumps(payload))
        return payload.request_id

    def poll(
        self, pattern: re.Pattern | None = None, block: bool = False
    ) -> Payload:
        payloads = self.poll_batch(pattern=pattern, block=block)
        for p in payloads[1:]:
            self._response_buffer[p.request_id] = p
        return payloads[0]

    def poll_batch(
        self, pattern: re.Pattern | None = None, block: bool = False
    ) -> List[Payload]:
        """Collect responses that match some pattern from the stream.

        This function may NOT actually pull from the stream. It may fetch something
        from the buffer, which records mismatched responses.

        Args:
            pattern (Optional[re.Pattern], optional): Only responses with this
                specific regex pattern will be returned.
                None means no pattern specified. Defaults to None.
            block (bool, optional): Whether to block to receive a
                response (with the given pattern). Defaults to False.
        """
        if not block:
            return self._poll_batch_nonblock(pattern)
        else:
            while True:
                try:
                    return self._poll_batch_nonblock(pattern)
                except NoMessage:
                    time.sleep(0.05)

    def _poll_batch_nonblock(
        self, pattern: Optional[re.Pattern] = None
    ) -> List[Payload]:
        # Check whether there's response in the buffer.
        # If so, return immediately.
        if pattern is None:
            pattern = re.compile(".*")

        payloads = []
        for req_id, p in self._response_buffer.items():
            if pattern.match(str(req_id)):
                payloads.append(p)
        for p in payloads:
            self._response_buffer.pop(p.request_id)
        if len(payloads) > 0:
            return payloads

        # Otherwise, pull from the socket.
        try:
            p_bytes = self.recv_socket.recv(flags=zmq.NOBLOCK)
        except zmq.ZMQError:
            raise NoMessage()
        payload: Payload = pickle.loads(p_bytes)
        # logger.info(f"Payload transfer time: {time.monotonic() - payload.send_time:.4f}s")
        self._response_buffer[payload.request_id] = payload

        payloads = []
        for req_id, p in self._response_buffer.items():
            if pattern.match(str(req_id)):
                payloads.append(p)
        for p in payloads:
            self._response_buffer.pop(p.request_id)
        if len(payloads) > 0:
            return payloads
        raise NoMessage()


class NameResolvingReplyServer:

    def __init__(
        self,
        experiment_name: str,
        trial_name: str,
        idx: int,
    ):
        self.context = zmq.Context.instance(io_threads=ZMQ_IO_THREADS)

        send_name = names.request_reply_stream(
            experiment_name, trial_name, "master_recv"
        )
        try:
            master_recv_addr = name_resolve.wait(send_name, timeout=300)
        except TimeoutError as e:
            logger.error(f"Worker timeout waiting for master receive stream.")
            raise e

        recv_name = names.request_reply_stream(
            experiment_name, trial_name, f"master_send_{idx}"
        )
        try:
            master_send_addr = name_resolve.wait(recv_name, timeout=300)
        except TimeoutError as e:
            logger.error(f"Worker timeout waiting for master send stream")
            raise e

        self.accept(master_send_addr, master_recv_addr)

        name_resolve.add_subentry(
            name=names.request_reply_stream(
                experiment_name, trial_name, PUBSUB_BARRIER_NAME
            ),
            value=socket.gethostbyname(socket.gethostname()),
            keepalive_ttl=60,
        )

    def accept(self, server_send_addr: str, server_recv_addr: str):
        recv_socket: zmq.Socket = self.context.socket(zmq.PULL)
        recv_socket.connect(f"tcp://{server_send_addr}")
        recv_socket.setsockopt(zmq.LINGER, 0)
        self.recv_socket = recv_socket

        send_socket: zmq.Socket = self.context.socket(zmq.PUSH)
        send_socket.connect(f"tcp://{server_recv_addr}")
        send_socket.setsockopt(zmq.LINGER, 0)
        self.send_socket = send_socket

    def post(self, payload: Payload) -> uuid.UUID:
        assert (
            payload.request_id is not None and payload.handle_name is not None
        )
        payload.send_time = time.monotonic()
        self.send_socket.send(pickle.dumps(payload))
        return payload.request_id

    def poll(self, block: bool = False) -> Payload:
        try:
            payload_bytes = self.recv_socket.recv(
                flags=0 if block else zmq.NOBLOCK
            )
        except zmq.ZMQError:
            raise NoMessage()

        payload: Payload = pickle.loads(payload_bytes)
        # logger.debug(f"Payload transfer time: {time.monotonic() - payload.send_time:.4f}s")
        return payload

    def close(self):
        self.recv_socket.close()
        self.send_socket.close()
        self.context.destroy()

    def __del__(self):
        self.close()


def make_master_stream(
    worker_info: system_api.WorkerInformation,
    n_subscribers: int,
    handler_routing: Dict[str | system_api.ModelShardID, int],
) -> NameResolvingRequstClient:
    return NameResolvingRequstClient(
        experiment_name=worker_info.experiment_name,
        trial_name=worker_info.trial_name,
        n_subscribers=n_subscribers,
        handler_routing=handler_routing,
    )


def make_worker_stream(
    worker_info: system_api.WorkerInformation,
    idx: int,
) -> NameResolvingReplyServer:
    return NameResolvingReplyServer(
        experiment_name=worker_info.experiment_name,
        trial_name=worker_info.trial_name,
        idx=idx,
    )



from collections import defaultdict
from typing import *
import dataclasses
import itertools
import json
import time

import torch
import torch.distributed

from realhf.api.core.config import ModelFamily, ModelName
from realhf.api.core.model_api import ReaLModelConfig
import realhf.api.core.system_api as system_api
import realhf.base.constants as constants
import realhf.base.topology as topology


def bcast_cost(
    param_size: float, bw: float, src: int, dsts: List[int], n_nodes_per_gpu=8
):
    src_node = src // n_nodes_per_gpu
    dst_nodes = [dst // n_nodes_per_gpu for dst in dsts]
    if src_node == dst_nodes[0] and all(
        dst_node == dst_nodes[0] for dst_node in dst_nodes
    ):
        return param_size * 2 * 8 / (1800 * 1024**3)
        # return 0.0
    else:
        # param size is in float16, bw is in Gbps
        return param_size * 2 * 8 / (bw * 1024**3) * len(set(dst_nodes))


def compute_cost(
    world_size: int,
    from_model_name: ModelName,
    to_model_name: ModelName,
    from_topo: topology.PipeModelDataParallelTopology,
    to_topo: topology.PipeModelDataParallelTopology,
    model_config: ReaLModelConfig,
    bw: float,  # Gbps
    set_interval_cost: float,
) -> int:
    from realhf.impl.model.comm.param_realloc import (
        _create_param_realloc_groups,
        _derive_reparallelize_comm_plan,
        ParamReallocInfo,
        ReparallelizeReceiverStep,
        ReparallelizeSenderStep,
    )

    param_sync_groups = {}
    param_sync_src_ranks = {}
    param_sync_dst_ranks = {}
    msid2mwid = {}
    for i in range(from_topo.world_size()):
        msid2mwid[
            system_api.ModelShardID.from_parallelism_rank(
                from_model_name, from_topo, i
            )
        ] = i
    for i in range(to_topo.world_size()):
        msid2mwid[
            system_api.ModelShardID.from_parallelism_rank(
                to_model_name, to_topo, i
            )
        ] = (i + world_size - to_topo.world_size())
    _create_param_realloc_groups(
        from_topo,
        to_topo,
        from_model_name,
        to_model_name,
        msid2mwid,
        param_sync_groups,
        param_sync_src_ranks,
        param_sync_dst_ranks,
    )
    pg_info = ParamReallocInfo(
        param_sync_groups,
        param_sync_src_ranks,
        param_sync_dst_ranks,
    )
    comm_plan = _derive_reparallelize_comm_plan(
        from_model_name,
        to_model_name,
        from_topo,
        to_topo,
        model_config,
        model_config,
        pg_info,
    )

    # Run boradcast!
    max_cost = max_comm_volume = max_bcast_cnt = 0
    for _rank in range(world_size):
        cost = comm_volume = bcast_cnt = 0
        for step in comm_plan:
            if (
                isinstance(step, ReparallelizeReceiverStep)
                and step.rank == _rank
            ):
                if step.rank != step.src:
                    cost += bcast_cost(
                        step.param_size, bw, step.src, step.dst_ranks
                    )
                    comm_volume += step.param_size
                    bcast_cnt += 1
                cost += set_interval_cost
            if isinstance(step, ReparallelizeSenderStep) and step.rank == _rank:
                if step.group is not None:
                    cost += bcast_cost(
                        step.param_size, bw, step.rank, step.dst_ranks
                    )
                    bcast_cnt += 1
        max_cost = max(max_cost, cost)
        max_comm_volume = max(max_comm_volume, comm_volume)
        max_bcast_cnt = max(max_bcast_cnt, bcast_cnt)

    return max_cost


def decompose_to_three_factors(n: int):
    factors = []
    for i in range(1, int(n ** (1 / 2)) + 1):
        if n % i == 0:
            for j in range(i, int((n // i) ** (1 / 2)) + 1):
                if (n // i) % j == 0:
                    k = (n // i) // j
                    factors += list(set(itertools.permutations([i, j, k])))
    return factors


def dump_table(
    n_nodes: int,
    model_family: ModelFamily,
    model_path: str,
    rank: int = 0,
    parallel: int = 1,
):
    from_model_name = ModelName("actor", 0)
    to_model_name = ModelName("actor", 1)

    def hash_tuple_into_str(t) -> str:
        return ",".join([str(i) for i in t])

    import tqdm

    res = {}
    device_mesh_sizes = [4] + [8 * i for i in range(1, n_nodes + 1)]
    space = list(itertools.product(device_mesh_sizes, device_mesh_sizes))
    sub_space = space[rank::parallel]
    # for a, b in set(small_spaces + large_spaces):
    for a, b in sub_space:
        mtik = time.perf_counter()
        all_configs = list(
            itertools.product(
                decompose_to_three_factors(a), decompose_to_three_factors(b)
            )
        )
        all_configs = list(
            filter(lambda x: x[0][1] <= 8 and x[1][1] <= 8, all_configs)
        )
        all_configs = list(
            filter(lambda x: x[0][2] <= 8 and x[1][2] <= 8, all_configs)
        )
        all_configs = list(
            filter(
                lambda x: x[0][1] in [1, 2, 4, 8] and x[1][1] in [1, 2, 4, 8],
                all_configs,
            )
        )
        all_configs = list(
            filter(lambda x: x[0][0] <= 16 and x[1][0] <= 16, all_configs)
        )
        all_configs = list(
            filter(
                lambda x: x[0][1] % x[1][1] == 0 or x[1][1] % x[0][1] == 0,
                all_configs,
            )
        )
        for config_id, (from_pp_mp_dp, to_pp_mp_dp) in tqdm.tqdm(
            enumerate(all_configs)
        ):
            world_size = max(a, b)

            from_topo = topology.PipeModelDataParallelTopology(
                *from_pp_mp_dp, False, False
            )
            to_topo = topology.PipeModelDataParallelTopology(
                *to_pp_mp_dp, False, False
            )
            assert world_size >= from_topo.world_size()
            assert world_size >= to_topo.world_size()

            from realhf.search_engine.utils import load_model_config

            mconfig = load_model_config(model_family._class, model_path)

            # tik = time.perf_counter()
            cost = compute_cost(
                world_size,
                from_model_name,
                to_model_name,
                from_topo,
                to_topo,
                mconfig,
                bw=200.0,
                set_interval_cost=0.03,
            )
            # print(
            #     f"Model size {model_size}: {from_pp_mp_dp} -> {to_pp_mp_dp}, cost {cost:.4f}",
            #     time.perf_counter() - tik,
            # )
            res[
                hash_tuple_into_str(
                    (model_family.size, *from_pp_mp_dp, *to_pp_mp_dp)
                )
            ] = int(cost * 1000 * 1000)
        print(
            f"Time for model size {model_family.size} {a} -> {b} {rank}/{parallel}: "
            f"{time.perf_counter() - mtik:.4f}, num res entries {len(res)}"
        )

    print(f"Rank {rank} of model {model_family} finished, res size {len(res)}.")

    import os
    import pickle

    dump_path = os.path.join(constants.PROFILER_CACHE_PATH, "param_realloc")
    fn = f"prtc_{model_family}_n{n_nodes}_{rank}_{parallel}.pkl"
    if not os.path.exists(dump_path):
        os.makedirs(dump_path, exist_ok=True)
    with open(os.path.join(dump_path, fn), "wb") as f:
        pickle.dump(res, f)
    print(
        f"dumped table with {len(res)} entries to {model_family}-{rank}-{parallel}."
    )


def dump_table_parallel(
    n_nodes: int,
    model_family_to_path: Dict[ModelFamily, str],
    parallel: int = 4,
):
    import torch.multiprocessing as mp

    mp.set_start_method("spawn", force=True)

    rq = mp.Queue()
    ps = []
    for model_family, model_path in model_family_to_path.items():
        for rank in range(parallel):
            ps.append(
                mp.Process(
                    target=dump_table,
                    args=(n_nodes, model_family, model_path, rank, parallel),
                )
            )

    for p in ps:
        p.start()

    for p in ps:
        p.join()


def merge_tables(
    n_nodes: int,
    model_family_to_path: Dict[ModelFamily, str],
    parallel: int = 4,
):
    import os
    import pickle

    res_path = os.path.join(constants.PROFILER_CACHE_PATH, "param_realloc")

    for model_family in model_family_to_path.keys():
        prefix = f"prtc_{model_family}_n{n_nodes}"

        r = {}
        counter = 0
        for fn in os.listdir(res_path):
            if fn.endswith(".pkl") and fn.startswith(prefix):
                counter += 1
                path = os.path.join(res_path, fn)
                with open(path, "rb") as f:
                    r.update(pickle.load(f))
                os.remove(path)
        if counter < parallel:
            raise RuntimeError(
                "missing sub-tables, probably some sub-processes failed "
                "during param realloc time cost estimation."
            )
        with open(os.path.join(res_path, f"{prefix}.pkl"), "wb") as f:
            pickle.dump(r, f)
        print(
            f"merged parallel tables into {prefix}.pkl, total entries {len(r)}"
        )


def estimate_param_realloc_time_cost(
    n_nodes: int,
    model_family_to_path: Dict[ModelFamily, str],
    parallel: int = 4,
):
    dump_table_parallel(n_nodes, model_family_to_path, parallel)
    merge_tables(n_nodes, model_family_to_path, parallel)



import argparse
import itertools
import time

import realhf.base.testing as testing

BATCH_SIZE_RANGE = [1, 2, 4, 8, 16, 32, 64, 128]
SEQ_LEN_RANGE = [128, 256, 512]


def profile_layer_func(
    world_size,
    model_path,
    model_name,
    warm_up_rounds,
    profile_rounds,
    batch_size_range,
    seq_len_range,
    use_sequence_parallel=False,
    use_gradient_checkpointing=False,
):
    # FIXME: use_sequence_parallel=True and use_gradient_checkpointing=True will cause bugs
    import torch

    import realhf.base.constants as constants

    testing.init_global_constants(
        1, world_size, 1, sequence_parallel=False, gradient_checkpointing=False
    )
    device = torch.device("cuda")
    with constants.model_scope(testing.MODEL_NAME):
        from realhf.search_engine.layers import make_profile_layers

        profile_layers = make_profile_layers(device, model_path, model_name)

        st = time.monotonic_ns()
        for i in range(warm_up_rounds + profile_rounds):
            for bs, seq_len in itertools.product(
                batch_size_range, seq_len_range
            ):
                profile_layers.fwd_gen(bs, seq_len)
                profile_layers.fwd_bwd_opt(bs, seq_len)

            if i < warm_up_rounds:
                profile_layers.reset_stats()
        profile_layers.make_dataframe_and_print()
        profile_layers.dump_stats(world_size)
        t = (time.monotonic_ns() - st) / int(1e9)
        print(f"profile world size {world_size} cost {t:4f} seconds")


if __name__ == "__main__":
    st = time.monotonic_ns()
    parser = argparse.ArgumentParser(prog="profile_layers")
    parser.add_argument(
        "--model_path",
        type=str,
        required=True,
    )
    parser.add_argument("--expr_name", type=str, default="profile")
    parser.add_argument("--trial_name", type=str, default="profile")
    parser.add_argument("--model_name", type=str, default="Llama-2-70b")
    parser.add_argument("--warm_up_rounds", type=int, default=1)
    parser.add_argument("--profile_rounds", type=int, default=3)
    # parser.add_argument("--use_sequence_parallel", action="store_true")
    # parser.add_argument("--use_gradient_checkpointing", action="store_true")
    args = parser.parse_args()

    world_sizes = [1, 2, 4, 8]

    for world_size in world_sizes:
        testing.clear_name_resolve(args.expr_name, args.trial_name)
        mp = testing.LocalMultiProcessTest(
            world_size,
            profile_layer_func,
            world_size,
            args.model_path,
            args.model_name,
            args.warm_up_rounds,
            args.profile_rounds,
            BATCH_SIZE_RANGE,
            SEQ_LEN_RANGE,
            expr_name=args.expr_name,
            trial_name=args.trial_name,
        )
        mp.launch()

    t = (time.monotonic_ns() - st) / int(1e9)
    print(f"profile model {args.model_name} time cost {t:4f} seconds")



from abc import ABC
from typing import Callable, List
import dataclasses
import math
import threading
import time

INFINITE_DURATION = 60 * 60 * 24 * 365 * 1000


class FrequencyControl:
    """An utility to control the execution of code with a time or/and step frequency."""

    def __init__(
        self, frequency_seconds=None, frequency_steps=None, initial_value=False
    ):
        """Initialization method of FrequencyControl.
        Args:
            frequency_seconds: Minimal interval between two trigger.
            frequency_steps: Minimal number of steps between two triggers.
            initial_value: In true, the first call of check() returns True.

        NOTE:
            - If both frequency_seconds and frequency_steps are None, the checking will always return False except
             for the specified initial value.
            - If passed both, both frequency and steps conditions have to be met for check() to return True.
            - If one is passed, checking on the other condition will be ignored.
        """
        self.frequency_seconds = frequency_seconds
        self.frequency_steps = frequency_steps
        self.__start_time = time.monotonic()
        self.__steps = 0
        self.__last_time = time.monotonic()
        self.__last_steps = 0
        self.__interval_seconds = self.__interval_steps = None
        self.__initial_value = initial_value
        self.__lock = threading.Lock()

    @property
    def total_seconds(self):
        return time.monotonic() - self.__start_time

    @property
    def total_steps(self):
        return self.__steps

    @property
    def interval_seconds(self):
        return self.__interval_seconds

    @property
    def interval_steps(self):
        return self.__interval_steps

    def check(self, steps=1):
        """Check whether frequency condition is met.
        Args:
            steps: number of step between this and the last call of check()

        Returns:
            flag: True if condition is met, False other wise
        """
        with self.__lock:
            now = time.monotonic()
            self.__steps += steps

            if self.__initial_value:
                self.__last_time = now
                self.__last_steps = self.__steps
                self.__initial_value = False
                return True

            self.__interval_seconds = now - self.__last_time
            self.__interval_steps = self.__steps - self.__last_steps
            if self.frequency_steps is None and self.frequency_seconds is None:
                return False
            if (
                self.frequency_seconds is not None
                and self.__interval_seconds < self.frequency_seconds
            ):
                return False
            if (
                self.frequency_steps is not None
                and self.__interval_steps < self.frequency_steps
            ):
                return False
            self.__last_time = now
            self.__last_steps = self.__steps

            return True

    def reset_time(self):
        self.__last_time = time.monotonic()


@dataclasses.dataclass
class EpochStepTimeFreqCtl:
    freq_epoch: int
    freq_step: int
    freq_sec: int

    def __post_init__(self):
        self.epoch_ctl = FrequencyControl(frequency_steps=self.freq_epoch)
        self.step_ctl = FrequencyControl(frequency_steps=self.freq_step)
        self.time_ctl = FrequencyControl(frequency_seconds=self.freq_sec)

    def check(self, epochs: int, steps: int):
        x, y, z = (
            self.epoch_ctl.check(epochs),
            self.step_ctl.check(steps),
            self.time_ctl.check(),
        )
        return x or y or z


@dataclasses.dataclass
class Scheduler(ABC):
    init_value: float
    total_iters: int

    def __post_init__(self):
        if self.total_iters <= 0:
            raise ValueError("total_iters should be a positive number.")

    def get(self, step: int) -> float:
        """Get the scheduled value at the current `step`."""
        if step < 0 or step > self.total_iters:
            raise ValueError(
                f"Scheduler step should be in the interval [0, {self.total_iters}]. Input {step}."
            )
        return self._get(step)

    def _get(self, step):
        raise NotImplementedError()

    @property
    def final_value(self):
        return self.get(step=self.total_iters)


@dataclasses.dataclass
class ConstantScheduler(Scheduler):

    def _get(self, *args, **kwargs) -> float:
        return self.init_value


@dataclasses.dataclass
class LinearScheduler(Scheduler):
    end_value: float

    def _get(self, step: int) -> float:
        return (
            self.end_value - self.init_value
        ) / self.total_iters * step + self.init_value


@dataclasses.dataclass
class ExponentialScheduler(Scheduler):
    decay: float

    def _get(self, step: int) -> float:
        return self.init_value * self.decay**step


@dataclasses.dataclass
class CosineDecayScheduler(Scheduler):
    end_value: float

    def __post_init__(self):
        super().__post_init__()
        if self.end_value >= self.init_value:
            raise ValueError("end_value should be smaller than init_value!")

    def _get(self, step: int) -> float:
        delta = self.init_value - self.end_value
        return (
            delta * 0.5 * (1 + math.cos(math.pi / self.total_iters * step))
            + self.end_value
        )


@dataclasses.dataclass
class ChainedScheduler:
    schedulers: List[Scheduler]

    @property
    def total_iters(self):
        return sum(x.total_iters for x in self.schedulers)

    @property
    def init_value(self):
        return self.schedulers[0].init_value

    @property
    def final_value(self):
        return self.schedulers[-1].final_value

    def __post_init__(self):
        for i in range(len(self.schedulers) - 1):
            # Float point err 1e-8.
            if (
                abs(
                    self.schedulers[i + 1].get(0)
                    - self.schedulers[i].final_value
                )
                > 1e-8
            ):
                raise ValueError(
                    f"Values should be consecutive between "
                    f"the {i}-th ({type(self.schedulers[i])}) and "
                    f"the {i+1}-th {type(self.schedulers[i+1])} schedulers! "
                    f"End value is {self.schedulers[i].final_value} and the "
                    f"next init value is {self.schedulers[i + 1].get(0)}."
                )

    def get(self, step: int) -> float:
        for s in self.schedulers:
            if step > s.total_iters:
                step -= s.total_iters
            else:
                return s.get(step)



from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, TYPE_CHECKING, Union
import abc
import copy
import dataclasses
import keyword
import os

import torch
import torch.utils.data
import transformers

from realhf.api.core import dfg, system_api
from realhf.api.core.config import ModelFamily, ModelName
from realhf.base.namedarray import NamedArray
import realhf.base.logging as logging

logger = logging.getLogger("model")


@dataclasses.dataclass
class ReaLModelConfig:
    """Configuration for ReaLModel.

    :param n_layers: Number of transformer blocks.
    :type n_layers: int
    :param n_kv_heads: Number of key-value attention heads.
    :type n_kv_heads: int
    :param head_dim: Dimension of each attention head.
        The number of query heads is hidden_dim // head_dim.
    :type head_dim: int
    :param hidden_dim: Hidden dimension of the transformer block.
    :type hidden_dim: int
    :param intermediate_dim: Dimension of the intermediate layer in the MLP.
    :type intermediate_dim: int
    :param vocab_size: Vocabulary size.
    :type vocab_size: int
    :param n_positions: Maximum context length. Can be None for
        rotary embedding, where the context length is decided during runtime.
    :type n_positions: Optional[int]
    :param embd_pdrop: Dropout probability for the embedding layer.
    :type embd_pdrop: float
    :param resid_pdrop: Dropout probability for the residual connections.
    :type resid_pdrop: float
    :param attn_pdrop: Dropout probability for the attention weights.
    :type attn_pdrop: float
    :param layer_norm_epsilon: Epsilon for layer normalization.
    :type layer_norm_epsilon: float
    :param activation_function: Activation function for the MLP.
    :type activation_function: str
    :param scale_attn_by_inverse_layer_idx: Whether to scale the attention weights
        by the inverse of the layer index.
    :type scale_attn_by_inverse_layer_idx: bool
    :param use_attention_bias: Whether to use attention bias.
    :type use_attention_bias: bool
    :param layer_norm_type: Type of layer normalization. Either None or "rms".
    :type layer_norm_type: Optional[str]
    :param mlp_type: Type of the MLP. Either None or "llama".
    :type mlp_type: Optional[str]
    :param apply_rotary: Whether to apply rotary embedding.
    :type apply_rotary: bool
    :param rotary_base: Exponential base for the rotary embedding.
    :type rotary_base: float
    :param rotary_interleaved: Whether to use interleaved rotary embedding.
    :type rotary_interleaved: bool
    :param rotary_scaling: Scaling factor for the rotary embedding.
    :type rotary_scaling: Optional[float]
    :param rotary_scaling_type: Type of scaling for the rotary embedding.
    :type rotary_scaling_type: Optional[str]
    :param is_critic: Whether the model is a critic model.
    :type is_critic: bool
    :param gradient_accumulation_fusion: Whether to fuse
        gradient accumulation in Megatron.
        Currently not supported.
    :type gradient_accumulation_fusion: bool
    :param share_embeddings_and_output_weights: Whether to share
        the embeddings and output weights.
        Currently not supported.
    :type share_embeddings_and_output_weights: bool
    """

    ### Architectural configurations. ###
    n_layers: int
    n_kv_heads: int
    head_dim: int
    hidden_dim: int
    intermediate_dim: int  # for mlp, usually 4*h
    vocab_size: int
    n_positions: Optional[int] = None
    embd_pdrop: float = 0.1
    resid_pdrop: float = 0.1
    attn_pdrop: float = 0.1
    layer_norm_epsilon: float = 1e-5
    activation_function: str = "gelu"
    scale_attn_by_inverse_layer_idx: bool = True
    # llama does not use attention bias and uses special MLP/LayerNorm layers
    use_attention_bias: bool = True
    layer_norm_type: Optional[str] = None
    mlp_type: Optional[str] = None
    # rotary embedding
    apply_rotary: bool = False
    rotary_base: float = 10000.0
    rotary_interleaved: bool = False
    rotary_scaling: Optional[float] = None
    rotary_scaling_type: Optional[str] = None
    # Whether it is a critic/reward model that outputs scores.
    is_critic: bool = False

    ### Running configurations. ###
    gradient_accumulation_fusion: bool = False

    # Placeholder, not implemented
    share_embeddings_and_output_weights: bool = False

    def __post_init__(self):
        assert not self.share_embeddings_and_output_weights


def load_hf_tokenizer(
    model_name_or_path: str,
    fast_tokenizer=True,
    padding_side: Optional[str] = None,
) -> transformers.PreTrainedTokenizerFast:
    kwargs = {}
    if padding_side is not None:
        kwargs["padding_side"] = padding_side
    if os.path.exists(model_name_or_path):
        # Locally tokenizer loading has some issue, so we need to force download
        model_json = os.path.join(model_name_or_path, "config.json")
        if "codet5" in model_name_or_path:
            tokenizer = transformers.RobertaTokenizer.from_pretrained(
                model_name_or_path, fast_tokenizer=fast_tokenizer, **kwargs
            )
        if os.path.exists(model_json):
            tokenizer = transformers.AutoTokenizer.from_pretrained(
                model_name_or_path, fast_tokenizer=fast_tokenizer, **kwargs
            )
    else:
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            model_name_or_path, fast_tokenizer=fast_tokenizer, **kwargs
        )
    if tokenizer.pad_token_id is None:
        tokenizer.pad_token_id = tokenizer.eos_token_id
    return tokenizer


@dataclasses.dataclass
class ModelVersion:
    epoch: int = 0
    epoch_step: int = 0
    global_step: int = 0


@dataclasses.dataclass
class FinetuneSpec:
    total_train_epochs: int
    total_train_steps: int
    steps_per_epoch: int


@dataclasses.dataclass
class Model:
    name: ModelName
    module: torch.nn.Module
    tokenizer: transformers.PreTrainedTokenizerFast
    device: Union[str, torch.device]
    dtype: Optional[torch.dtype] = None
    version: ModelVersion = dataclasses.field(default_factory=ModelVersion)
    ft_spec: FinetuneSpec = None  # will be initialized by the backend

    def __post_init__(self):
        try:
            self.module = self.module.to(self.device)
        except ValueError as e:
            # 4-bit and 8-bit model may fail here
            logger.warning(
                f"Failed to move model to device {self.device} because {e}. Abort to device."
            )

    def inc_version(self):
        self.version.global_step += 1
        self.version.epoch_step += 1
        if self.version.epoch_step >= self.ft_spec.steps_per_epoch:
            self.version.epoch_step = 0
            self.version.epoch += 1


class ModelBackend(abc.ABC):

    @abc.abstractmethod
    def _initialize(self, model: Model, spec: FinetuneSpec) -> Model:
        raise NotImplementedError()

    def initialize(self, model: Model, spec: FinetuneSpec) -> Model:
        model.ft_spec = spec
        return self._initialize(model, spec)


class NullBackend(ModelBackend):

    def _initialize(self, model: Model, spec: FinetuneSpec) -> Model:
        return model


def null_model(name: ModelName, device: Union[str, torch.device]) -> Model:
    return Model(name, torch.nn.Identity(), None, device)


def tokenizer_only_model(
    name: ModelName, device: Union[str, torch.device], tokenizer_path: str
) -> Model:
    return Model(
        name, torch.nn.Identity(), load_hf_tokenizer(tokenizer_path), device
    )


class ModelInterface(abc.ABC):
    """Interface for model training, evaluation, inference and generation."""

    def save(self, model: Model, save_dir: str):
        pass

    def evaluate(
        self, model: Model, eval_dataloader: torch.utils.data.DataLoader
    ) -> Dict:
        return {}

    def inference(self, model: Model, data: NamedArray) -> NamedArray:
        raise NotImplementedError()

    def generate(self, model: Model, data: NamedArray) -> NamedArray:
        raise NotImplementedError()

    def train_step(self, model: Model, data: NamedArray) -> Dict:
        raise NotImplementedError()


ALL_MODEL_CLASSES = {}
ALL_INTERFACE_CLASSES = {}
ALL_BACKEND_CLASSES = {}
ALL_WRAPPER_CLASSES = {}


def register_model(name, model_cls):
    assert name not in ALL_MODEL_CLASSES
    ALL_MODEL_CLASSES[name] = model_cls


def register_interface(name, cls_):
    assert name not in ALL_INTERFACE_CLASSES
    assert issubclass(cls_, ModelInterface)
    ALL_INTERFACE_CLASSES[name] = cls_


def register_backend(name, cls_):
    assert name not in ALL_BACKEND_CLASSES
    assert issubclass(cls_, ModelBackend)
    ALL_BACKEND_CLASSES[name] = cls_


def register_wrapper(name, cls_):
    assert name not in ALL_WRAPPER_CLASSES
    ALL_WRAPPER_CLASSES[name] = cls_


def make_model_wrapper(
    cfg: system_api.ModelWrapper,
) -> Callable[[Model], Model]:
    cls_ = ALL_WRAPPER_CLASSES[cfg.type_]
    return cls_(**cfg.args)


def make_model(
    cfg: system_api.Model, name: ModelName, device: Union[str, torch.device]
) -> Model:
    logger.debug(f"making model {cfg.type_} on {device}")
    model_cls = ALL_MODEL_CLASSES[cfg.type_]
    model = model_cls(**cfg.args, name=name, device=device)
    assert isinstance(model, Model)
    for w in cfg.wrappers:
        model = make_model_wrapper(w)(model)
        assert isinstance(model, Model)
    return model


def make_interface(cfg: dfg.ModelInterface) -> ModelInterface:
    cls_ = ALL_INTERFACE_CLASSES[cfg.type_]
    return cls_(**cfg.args)


def make_backend(cfg: system_api.ModelBackend) -> ModelBackend:
    cls_ = ALL_BACKEND_CLASSES[cfg.type_]
    return cls_(**cfg.args)


register_backend("null", NullBackend)
register_model("null", null_model)
register_model("tokenizer", tokenizer_only_model)

SUPPORTED_MODELS = []
HF_MODEL_FAMILY_REGISTRY = {}


def is_valid_function_name(name):
    if not name.isidentifier():
        return False
    if keyword.iskeyword(name):
        return False
    return True


def register_hf_family(
    name: str,
    hf_cls_name: str,
    config_from_hf_converter: Callable[
        [transformers.PretrainedConfig], ReaLModelConfig
    ],
    config_to_hf_converter: Callable[
        [ReaLModelConfig], transformers.PretrainedConfig
    ],
    sd_from_hf_converter: Callable[[Dict, ReaLModelConfig], Dict],
    sd_to_hf_converter: Callable[[Dict, ReaLModelConfig], Dict],
    embedding_param_names: Callable[[ReaLModelConfig], List[str]],
    tblock_param_names: Callable[[ReaLModelConfig, int], List[str]],
    head_param_names: Callable[[ReaLModelConfig], List[str]],
):
    if name in SUPPORTED_MODELS:
        raise ValueError(f"Model {name} is already registered.")
    if not is_valid_function_name(name):
        raise ValueError(
            f"Model name {name} is not a valid function name in Python."
        )
    SUPPORTED_MODELS.append(name)
    HF_MODEL_FAMILY_REGISTRY[name] = dict(
        name=name,
        hf_cls_name=hf_cls_name,
        config_from_hf_converter=config_from_hf_converter,
        config_to_hf_converter=config_to_hf_converter,
        sd_from_hf_converter=sd_from_hf_converter,
        sd_to_hf_converter=sd_to_hf_converter,
        embedding_param_names=embedding_param_names,
        tblock_param_names=tblock_param_names,
        head_param_names=head_param_names,
    )



from realhf.api.core.model_api import ReaLModelConfig
import realhf.base.constants as constants
import realhf.base.namedarray


def random_sample(bs, seq_len, vocab_size):
    import torch

    from realhf.impl.model.utils.padding import unpad_input

    input_ids = torch.randint(0, vocab_size, (bs, seq_len), dtype=torch.long)
    attention_mask = torch.ones_like(input_ids)
    packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
        input_ids, attention_mask
    )
    prompt_mask = torch.zeros_like(packed_input_ids)
    data = realhf.base.namedarray.NamedArray(
        packed_input_ids=packed_input_ids,
        cu_seqlens=cu_seqlens,
        # prompts=input_ids,
        prompt_mask=prompt_mask.bool(),
        # prompt_att_mask=attention_mask,
    )
    seqlens_cpu = [seq_len for _ in range(bs)]
    data.register_metadata(seqlens=seqlens_cpu)
    return data


def find_factors(n):
    factors = []
    for i in range(1, n + 1):
        if n % i == 0:
            factors.append(i)
    return factors


def make_stats_key(rpc_name, bs, seq_len):
    return f"{rpc_name}|{bs}|{seq_len}"


def parse_stats_key(key):
    rpc_name, bs, seq_len = key.split("|")
    return rpc_name, int(bs), int(seq_len)


def load_model_config(model_class: str, model_path: str) -> ReaLModelConfig:
    from realhf.impl.model.nn.real_llm_api import ReaLModel

    return getattr(ReaLModel, f"config_from_{model_class}")(
        model_path=model_path
    )



from typing import Dict, List, Optional
import argparse
import dataclasses

import colorama
import torch
import transformers

from realhf.api.core import config as config_api
from realhf.api.core import dfg, model_api, system_api
from realhf.api.core.system_api import ExperimentConfig
from realhf.api.quickstart.entrypoint import register_quickstart_exp
from realhf.apps.quickstart import main
from realhf.base import logging
from realhf.base.namedarray import NamedArray
from realhf.experiments.common.ppo_exp import PPOConfig

logger = logging.getLogger("Sentiment PPO example")


@dataclasses.dataclass
class SentimentScoringInterface(model_api.ModelInterface):

    def __post_init__(self):
        super().__post_init__()
        self.score_model = (
            transformers.AutoModelForSequenceClassification.from_pretrained(
                "/path/to/score_model"
            ).cuda()
        )
        self.score_model.eval()

        self.score_tokenizer = transformers.AutoTokenizer.from_pretrained(
            "/path/to/score_model"
        )

    @torch.no_grad()
    def inference(self, model: model_api.Model, data: NamedArray) -> NamedArray:
        packed_input_ids: torch.Tensor = data["packed_input_ids"]
        seqlens_cpu: List[int] = data.metadata["seqlens"]
        max_seqlen = max(seqlens_cpu)
        bs = len(seqlens_cpu)
        device = packed_input_ids.device

        # Build attention mask.
        _ind = torch.arange(max_seqlen, dtype=torch.long, device=device)
        _seqlens = torch.tensor(seqlens_cpu, dtype=torch.long, device=device)
        attention_mask = _ind.unsqueeze(0) < _seqlens.unsqueeze(1)

        # Pad input_ids.
        input_ids = torch.full(
            (bs * max_seqlen,),
            fill_value=model.tokenizer.pad_token_id,
            device=device,
            dtype=torch.long,
        )
        indices = torch.nonzero(
            attention_mask.flatten(), as_tuple=False
        ).flatten()
        input_ids[indices] = packed_input_ids
        input_ids = input_ids.view(bs, max_seqlen)

        # Re-tokenize.
        texts = model.tokenizer.batch_decode(
            input_ids, skip_special_tokens=True
        )
        encoding = self.score_tokenizer(
            texts, return_tensors="pt", padding=True, truncation=True
        )

        # Inference to get the score.
        logits = self.score_model(
            input_ids=encoding["input_ids"].cuda(),
            attention_mask=encoding["attention_mask"].cuda(),
        ).logits
        # For IMDB, 0 is negative and 1 is positive. We record the logits of positive.
        scores = logits[..., -1].contiguous().float()
        assert scores.shape == (bs,), scores.shape

        # ###################### logging ######################
        # for text, score in zip(texts, scores):
        #     logger.info(
        #         f"reward is {colorama.Fore.RED}{score.item()}{colorama.Style.RESET_ALL}, "
        #         f"sequence is: {colorama.Fore.YELLOW + colorama.Style.DIM}{text}{colorama.Style.RESET_ALL}"
        #     )
        # #####################################################

        res = NamedArray(scores=scores)
        res.register_metadata(**data.metadata)
        return res


model_api.register_interface("sentiment_scoring", SentimentScoringInterface)


class MyPPOConfig(PPOConfig):

    def initial_setup(self) -> ExperimentConfig:
        if (
            self.rew_inf.parallel.model_parallel_size > 1
            or self.rew_inf.parallel.pipeline_parallel_size > 1
        ):
            raise ValueError(
                "For this example, the reward model does not support model parallel or pipeline parallel."
            )

        cfg = super().initial_setup()

        # Replace the backend and model configurations for the reward model.
        for mw in cfg.model_worker:
            for s in mw.shards:
                if s.id.model_name.role == "reward":
                    s.model = config_api.Model(
                        "tokenizer",
                        args=dict(
                            tokenizer_path=self.rew.path,
                        ),
                    )
                    s.backend = config_api.ModelBackend("null")

        # Change the model function call implementation.
        idx = 0
        for rpc in cfg.model_rpcs:
            if rpc.model_name.role == "reward":
                break
            idx += 1
        inf_reward_rpc = cfg.model_rpcs[idx]
        inf_reward_rpc.interface_impl = dfg.ModelInterface("sentiment_scoring")
        inf_reward_rpc.post_hooks = []

        return cfg


register_quickstart_exp("my-ppo", MyPPOConfig)

if __name__ == "__main__":
    main()



from typing import List

from realhf.api.core.dfg import build_graph as build_dfg
from realhf.api.core.dfg import MFCDef, ModelInterfaceType
from realhf.api.quickstart.device_mesh import (
    DeviceMesh,
    find_parallel_strategies,
)
from realhf.api.quickstart.search import MFCDef, RPCExecution, RPCInstance
from realhf.search_engine.estimate import (
    estimate_rpc_memory_cost,
    estimate_rpc_time_cost,
)

MEM_INDEX = 1.0  # heuristic value to scale estimated memory


def enumerate_rpc_executions(
    rpc: MFCDef,
    device_mesh: DeviceMesh,
    seq_len: int,
    num_gen_tokens: int,
    n_ppo_minibatches: int,
    gradient_checkpointing: bool,
) -> List[RPCExecution]:
    sub_device_meshes = device_mesh.sub_device_meshes()
    import pprint

    feasible = []
    for sub_device_mesh in sub_device_meshes:
        ps = find_parallel_strategies(sub_device_mesh)
        for parallel in ps:
            num_dp = parallel.data_parallel_size
            num_pp = parallel.pipeline_parallel_size
            num_mp = parallel.model_parallel_size
            bs = rpc.min_n_seqs
            # seq_len = seq_len
            min_bs = (
                2 * num_dp * num_pp * n_ppo_minibatches
                if rpc.interface_type == ModelInterfaceType.TRAIN_STEP
                else num_dp * num_pp
            )
            if min_bs > bs:
                # batch size too small
                continue
            # heuristic to filter out inherent slow configurations
            if (
                num_mp * num_dp > device_mesh.n_gpus_per_node
                and rpc.interface_type == ModelInterfaceType.TRAIN_STEP
            ):
                continue
            if num_mp > 8:
                continue
            if num_pp > max(device_mesh.n_nodes, 8):
                continue
            # memory and time estimation
            mem_cost, static_mem = estimate_rpc_memory_cost(
                rpc,
                parallel,
                bs,
                seq_len,
                gradient_checkpointing=gradient_checkpointing,
                n_ppo_minibatches=n_ppo_minibatches,
                num_gen_tokens=num_gen_tokens,
                offload=rpc.model_name.role in ["ref", "reward"],
            )
            mem_cost = int(mem_cost * MEM_INDEX)
            static_mem = int(static_mem * MEM_INDEX)
            time_cost = estimate_rpc_time_cost(
                rpc,
                parallel,
                bs=bs,
                seq_len=seq_len,
                num_gen_tokens=num_gen_tokens,
                gradient_checkpointing=gradient_checkpointing,
                n_ppo_minibatches=n_ppo_minibatches,
            )
            time_cost = int(time_cost)
            if mem_cost < device_mesh.gpu_memory_capacity:
                feasible.append(
                    RPCExecution(
                        rpc,
                        sub_device_mesh,
                        parallel,
                        time_cost,
                        mem_cost,
                        static_mem,
                    )
                )
    return feasible


def build_graph(
    rpcs: List[MFCDef],
    num_epoch: int = 5,
    epoch_dependency_interval: int = 1,
    if_print=False,
) -> List[RPCInstance]:
    """Build model function call graph of multiple training epochs,

    args:
        exp: ProfileExperiment, the experiment object
        num_epoch: int, number of training epochs
        epoch_dependency_interval: int, the interval of epoch dependency,
            e.g. if epoch_dependency_interval = 2, then the graph will have
            edges between epoch i and epoch i+2, i+4, ...

    returns:
        rpc_instances: List[RPCInstance], the list of RPCInstance objects
    """
    # one epoch dependency graph
    rpcs, edges = build_dfg(rpcs)
    rpc_names_mapping = {rpc.name: rpc for rpc in rpcs}
    rpc_instances = []

    # multi epoch graph
    for epoch_id in range(num_epoch):
        for rpc in rpcs:
            children = []
            parents = []
            if rpc.is_src and epoch_id >= epoch_dependency_interval:
                for other in rpcs:
                    if (
                        other.is_dst
                        and other.model_name.role == rpc.model_name.role
                    ):
                        parents.append(
                            RPCInstance(
                                rpc,
                                epoch_id - epoch_dependency_interval,
                                [],
                                [],
                            )
                        )
            if rpc.is_dst and rpc.model_name.role == rpc.model_name.role:
                for other in rpcs:
                    if (
                        other.is_src
                        and epoch_id + epoch_dependency_interval < num_epoch
                    ):
                        children.append(
                            RPCInstance(
                                rpc,
                                epoch_id + epoch_dependency_interval,
                                [],
                                [],
                            )
                        )
            for parent in rpc.parents:
                p = rpc_names_mapping[parent]
                parents.append(RPCInstance(p, epoch_id, [], []))
            for child in rpc.children:
                c = rpc_names_mapping[child]
                children.append(RPCInstance(c, epoch_id, [], []))
            rpc_instance = RPCInstance(rpc, epoch_id, parents, children)
            rpc_instances.append(rpc_instance)
    if if_print:
        for ri in rpc_instances:
            print(ri)
    return rpc_instances



from typing import Tuple

import torch
import torch.nn.functional as F


def dpo_loss(
    pi_logps: torch.Tensor,
    ref_logps: torch.Tensor,
    beta: float,
):
    assert len(pi_logps.shape) == 1 and pi_logps.shape[0] % 2 == 0, (
        pi_logps.shape,
        ref_logps.shape,
    )
    assert len(ref_logps.shape) == 1 and ref_logps.shape[0] % 2 == 0, (
        pi_logps.shape,
        ref_logps.shape,
    )
    pi_logps = pi_logps.view(-1, 2)
    ref_logps = ref_logps.view(-1, 2)
    pi_yw_logps, pi_yl_logps = pi_logps[:, 0], pi_logps[:, 1]
    ref_yw_logps, ref_yl_logps = ref_logps[:, 0], ref_logps[:, 1]
    pi_logratios = pi_yw_logps - pi_yl_logps
    ref_logratios = ref_yw_logps - ref_yl_logps
    losses = -F.logsigmoid(beta * (pi_logratios - ref_logratios)).mean()
    pos_score = beta * (pi_yw_logps - ref_yw_logps).detach().sum()
    neg_score = beta * (pi_yl_logps - ref_yl_logps).detach().sum()
    kl = -(pi_logps - ref_logps).detach().sum()
    return losses, pos_score, neg_score, kl



from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import dataclasses
import functools

import numpy as np
import torch
import torch.distributed
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
import transformers

from realhf.api.core import model_api
from realhf.api.core.config import ModelName
from realhf.base import constants, logging, topology
from realhf.base.monitor import cuda_tmark, cuda_tmarked, CUDATimeMarkType
from realhf.impl.model.comm.global_comm import NCCLProcessGroupInfo
from realhf.impl.model.comm.param_realloc import (
    _derive_reparallelize_comm_plan,
    fetch_trainable_params,
    is_trainable,
    ReparallelizeReceiverStep,
    ReparallelizeSenderStep,
    ReparallelizeTraget,
    store_trainable_params,
)
from realhf.impl.model.nn.flatten_param import set_intervals, slice_intervals
from realhf.impl.model.utils.padding import pad_input, unpad_input

from .flatten_param import build_param_spec, map_param_to_contigous_memory
from .real_llm_base import (
    OutputHead,
    PipeCacheData,
    PipeTransferData,
    real_model_embed_param_count,
    real_model_head_param_count,
    real_model_tblock_param_count,
    ReaLModelBlock,
    SequenceParallelActorHead,
    SequenceParallelCriticHead,
    VocabPositionEmbedding,
)
from .real_llm_generate import generate, GenerationConfig
from .real_llm_parallel import partition_pipeline_layers

logger = logging.getLogger("ReaLModel Interface")


@dataclasses.dataclass
class DuckModelOutput:
    logits: Optional[Union[List[torch.Tensor], torch.Tensor]] = None


@dataclasses.dataclass
class DuckGenerationOutput:
    sequences: torch.Tensor
    scores: Optional[torch.Tensor] = None
    logits_mask: Optional[torch.Tensor] = None


class ReaLModel(nn.Module):
    """The transformer model used in ReaL.

    This model supports 3D parallelism, offloaded inference,
    and parameter reallocation. It is usually more efficient
    than HuggingFace implementations.

    During construction, model parameters are not instantiated
    immediately because the model may be redistributed.
    The method ``instantiate`` should be called before using
    model parameters, e.g., forward or state dict.

    :param config: The model configuration.
    :type config: model_api.ReaLModelConfig
    :param dtype: The data type of the model.
    :type dtype: Optional[torch.dtype], optional
    :param device: The device of the model.
    """

    def __init__(
        self,
        config: model_api.ReaLModelConfig,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        if dtype is None:
            dtype = torch.float16
        self.config = config
        self.dtype = dtype
        self.device = device

        # The main attribute of the model: layers,
        # including the embedding layer, decoder layers, and the output head.
        self.layer_mapping = partition_pipeline_layers(
            config,
            constants.pipe_parallel_world_size(),
            real_model_embed_param_count,
            real_model_tblock_param_count,
            real_model_head_param_count,
        )
        self.layer_idx_start = self.layer_mapping[
            constants.pipe_parallel_rank()
        ][0]
        self.layer_idx_end = self.layer_mapping[constants.pipe_parallel_rank()][
            1
        ]
        self.num_stages = constants.pipe_parallel_world_size()

        self.layers = nn.ModuleList()

        # The model is lazily instantiated due to parameter reallocation.
        # For models that will be redistributed, we instantiate replica 0
        # and do not instantiate other replicas.
        self._instantiated = False
        self._instantiation_hooks = []

        # Attributes used for parameter reallocation.
        self._reparallelize_targets: Dict[
            Tuple[ModelName, ModelName], ReparallelizeTraget
        ] = {}

        # Attributes used for offload.
        self._offload_buffer = None
        self._offloaded = False

        # Attributes used for flattening parameters.
        self._param_spec, self._param_size = build_param_spec(
            list(range(self.layer_idx_start, self.layer_idx_end)),
            self.config,
            constants.model_parallel_world_size(),
            constants.sequence_parallel(),
        )
        self.contiguous_param = None

    @property
    def pre_process(self):
        # A workaround to make Megatron-LM backend happy.
        if constants.pipe_parallel_rank() == 0:
            return self.layers[0]
        elif (
            constants.pipe_parallel_rank()
            == constants.pipe_parallel_world_size() - 1
        ):
            return self.layers[-1]
        return None

    @property
    def post_process(self):
        # A workaround to make Megatron-LM backend happy.
        if (
            constants.pipe_parallel_rank()
            == constants.pipe_parallel_world_size() - 1
        ):
            return self.layers[-1]
        return None

    @property
    def share_embeddings_and_output_weights(self):
        return self.config.share_embeddings_and_output_weights

    def instantiate(self):
        """Instantiate the model parameters.

        Note that users can append hooks to this method
        to do more processing, such as loading from HuggingFace models.
        """
        assert not self._instantiated
        layers = []
        for idx in range(self.layer_idx_start, self.layer_idx_end):
            layers.append(self._build_layer(idx, self.config))

        self.layers = nn.ModuleList(layers)

        self.contiguous_param = torch.empty(
            self._param_size, dtype=self.dtype, device=self.device
        )
        map_param_to_contigous_memory(
            self.layers,
            self._param_spec,
            self.contiguous_param,
            self.layer_idx_start,
        )

        for h in self._instantiation_hooks:
            h()

        self._instantiated = True
        self._instantiation_hooks = []

    @property
    def num_layers(self):
        """Return the number of embedding or transformer layers in this pipeline stage."""
        return self.layer_idx_end - self.layer_idx_start

    @property
    def is_critic(self):
        return self.config.is_critic

    def _build_layer(
        self, idx: int, config: model_api.ReaLModelConfig
    ) -> nn.Module:
        dtype = self.dtype
        device = self.device
        if idx == 0:
            l = VocabPositionEmbedding(config, dtype=dtype, device=device)
        elif idx == config.n_layers + 1:
            l = self._build_output_head(config)
        else:
            l = ReaLModelBlock(
                config=config,
                layer_index=idx - 1,
                output_layernorm=(idx == config.n_layers),
                dtype=dtype,
                device=device,
            )
        return l

    def _build_output_head(
        self, config: model_api.ReaLModelConfig
    ) -> nn.Module:
        dtype = self.dtype
        device = self.device
        if config.is_critic and constants.sequence_parallel():
            l = SequenceParallelCriticHead(
                config.hidden_dim,
                1,
                bias=False,
                device=device,
                dtype=dtype,
            )
        elif not config.is_critic and constants.model_parallel_world_size() > 1:
            l = SequenceParallelActorHead(
                config.hidden_dim,
                config.vocab_size,
                bias=False,
                async_tensor_model_parallel_allreduce=not constants.sequence_parallel(),
                gradient_accumulation_fusion=config.gradient_accumulation_fusion,
                device=device,
                dtype=dtype,
            )
        else:
            l = OutputHead(
                config.hidden_dim,
                1 if config.is_critic else config.vocab_size,
                bias=False,
                device=device,
                dtype=dtype,
            )
        return l

    def async_offload(self):
        """Trigger offload asynchronously."""
        assert not self._offloaded
        assert self._instantiated
        if self._offload_buffer is None:
            self._offload_buffer = torch.empty_like(
                self.contiguous_param,
                dtype=self.dtype,
                device="cpu",
                pin_memory=True,
            )
        else:
            assert self._offload_buffer.shape == self.contiguous_param.shape
        dummy_tensor = torch.tensor((), device=self.device, dtype=self.dtype)
        self._offload_stream = torch.cuda.Stream()
        self._offload_event = torch.cuda.Event()
        self.contiguous_param = None
        for i, l in enumerate(self.layers):
            layer_idx = self.layer_idx_start + i
            with torch.cuda.stream(self._offload_stream):
                for k, p in l.named_parameters():
                    spec = self._param_spec[f"{layer_idx}.{k}"]
                    self._offload_buffer[spec.start_idx : spec.end_idx].copy_(
                        p.data.view(-1), non_blocking=True
                    )
                    p.data = dummy_tensor
        self._offload_event.record(self._offload_stream)
        self._offloaded = True

    def wait_for_offload(self):
        """Wait for offload to finish."""
        assert self._offloaded
        torch.cuda.current_stream().wait_event(self._offload_event)

    def __overlapped_load_forward(
        self, x: PipeTransferData, ys: List[PipeCacheData]
    ) -> Tuple[PipeTransferData, List[PipeCacheData]]:
        assert len(ys) == self.num_layers
        raw_pp_input = x.pp_input
        self.contiguous_param = torch.empty(
            self._param_size, dtype=self.dtype, device=self.device
        )
        map_param_to_contigous_memory(
            self.layers,
            self._param_spec,
            self.contiguous_param,
            self.layer_idx_start,
        )
        self.wait_for_offload()

        stream = torch.cuda.Stream()
        events: List[torch.cuda.Event] = [
            torch.cuda.Event() for _ in range(self.num_layers)
        ]
        with torch.cuda.stream(stream):
            for layer_idx, y, l, e in zip(
                range(self.layer_idx_start, self.layer_idx_end),
                ys,
                self.layers,
                events,
            ):
                # NOTE: although we can do more fine-grained overlapping, the overhead that can be
                # reduced is very small (~50ms), which is unnecessary for now.
                for k, v in l.named_parameters():
                    spec = self._param_spec[f"{layer_idx}.{k}"]
                    v.data.copy_(
                        self._offload_buffer[
                            spec.start_idx : spec.end_idx
                        ].view(spec.shape),
                        non_blocking=True,
                    )
                e: torch.cuda.Event
                e.record(stream)

        for layer_idx, y, l, e in zip(
            range(self.layer_idx_start, self.layer_idx_end),
            ys,
            self.layers,
            events,
        ):
            torch.cuda.default_stream().wait_event(e)
            x = l(x, y)
            x.pp_input = x.pp_output
        self._offloaded = False
        x.pp_input = raw_pp_input
        return x, ys

    def __forward(
        self, x: PipeTransferData, ys: List[PipeCacheData]
    ) -> Tuple[PipeTransferData, List[PipeCacheData]]:
        layers = self.layers
        assert len(ys) == len(layers), (len(ys), len(layers))
        raw_pp_input = x.pp_input
        for i, (layer, y) in enumerate(zip(layers, ys)):
            x = layer(x, y)
            x.pp_input = x.pp_output
        # Finally, pp_input is the input of this pipeline stage (maybe across several layers),
        # pp_output is the output of this pipeline stage.
        # In the first stage, pp_input is None.
        x.pp_input = raw_pp_input
        return x, ys

    def forward(
        self, x: PipeTransferData, ys: List[PipeCacheData]
    ) -> Tuple[PipeTransferData, List[PipeCacheData]]:
        if x.max_seqlen is not None and not isinstance(x.max_seqlen, int):
            x.max_seqlen = int(x.max_seqlen)
        if x.cu_seqlens is not None and not isinstance(
            x.cu_seqlens, torch.IntTensor
        ):
            x.cu_seqlens = x.cu_seqlens.int()

        # Copy input tensor to a pinned buffer.
        mp_size = constants.model_parallel_world_size()
        batch_length = None
        if ys[0].packed_input_ids is not None:
            batch_length = ys[0].packed_input_ids.shape[0]
        if x.pp_input is not None:
            batch_length = x.pp_input.shape[0]
        assert batch_length is not None
        padded_batch_length = (batch_length + mp_size - 1) // mp_size * mp_size
        pad_size = padded_batch_length - batch_length

        if (
            constants.sequence_parallel()
            and pad_size > 0
            and ys[0].packed_input_ids is not None
        ):
            _cu_seqlens = x.cu_seqlens
            _max_seqlen = x.max_seqlen
            _input_ids = ys[0].packed_input_ids
            _pp_input = x.pp_input

            x.cu_seqlens = torch.nn.functional.pad(
                x.cu_seqlens, (0, 1), value=padded_batch_length
            )
            x.max_seqlen = max(x.max_seqlen, padded_batch_length - batch_length)
            if ys[0].packed_input_ids is not None:
                input_ids_buf = torch.zeros(
                    (padded_batch_length,),
                    dtype=torch.long,
                    device=self.device,
                )
                input_ids_buf[:batch_length] = ys[0].packed_input_ids
                ys[0].packed_input_ids = input_ids_buf

            if x.pp_input is not None:
                pp_input_buf = torch.zeros(
                    (padded_batch_length, *x.pp_input.shape[1:]),
                    dtype=x.pp_input.dtype,
                    device=self.device,
                )
                pp_input_buf[:batch_length] = x.pp_input
                x.pp_input = pp_input_buf

        tmark_type = CUDATimeMarkType.forward
        with cuda_tmarked("fwd", tmark_type):
            # Main forward calls.
            if not self._offloaded:
                x, ys = self.__forward(x, ys)
            else:
                x, ys = self.__overlapped_load_forward(x, ys)

        # Resume from padding.
        if (
            constants.sequence_parallel()
            and pad_size > 0
            and ys[0].packed_input_ids is not None
        ):
            x.pp_output = x.pp_output[:-pad_size]

            x.pp_input = _pp_input
            ys[0].packed_input_ids = _input_ids
            x.cu_seqlens = _cu_seqlens
            x.max_seqlen = _max_seqlen

            if x.store_kv_cache:
                for y in ys:
                    if y.k_cache is not None:
                        y.k_cache = y.k_cache[:-pad_size]
                    if y.v_cache is not None:
                        y.v_cache = y.v_cache[:-pad_size]

        # Release the memory used for TP gathering.
        constants.clear_global_memory_buffer()
        return x, ys

    def _forward(
        self,
        input_ids: torch.LongTensor,
        cu_seqlens: torch.IntTensor,
        position_ids: torch.LongTensor,
        hidden_states: Optional[torch.Tensor],
        k_caches: Optional[List[torch.Tensor]],
        v_caches: Optional[List[torch.Tensor]],
        cache_seqlens: Optional[torch.IntTensor],
        max_seqlen: Optional[int],
    ):
        if k_caches is None:
            assert v_caches is None
            assert cache_seqlens is None
            k_caches = [None] * self.num_layers
            v_caches = [None] * self.num_layers

        h = hidden_states
        for idx, l in enumerate(self.layers):
            if isinstance(l, VocabPositionEmbedding):
                h = l._forward(input_ids, position_ids)
            elif isinstance(l, ReaLModelBlock):
                h, _, _ = l._forward(
                    h,
                    cu_seqlens=cu_seqlens,
                    k_cache=k_caches[idx],
                    v_cache=v_caches[idx],
                    cache_seqlens=cache_seqlens,
                    max_seqlen=max_seqlen,
                    attention_mask=None,
                )
            elif isinstance(
                l,
                (
                    OutputHead,
                    SequenceParallelCriticHead,
                    SequenceParallelActorHead,
                ),
            ):
                h = l._forward(h)
            else:
                raise NotImplementedError(f"Unsupported layer type {type(l)}")

        return h

    def state_dict(self, *args, **kwargs):
        """Map layer indices to global layer indices."""
        state_dict = self.layers.state_dict(*args, **kwargs)
        new_state_dict = {}
        for k, v in state_dict.items():
            k = k.lstrip("module.").lstrip("layers.")
            local_idx = int(k.split(".")[0])
            name = k.split(".", 1)[1]
            new_state_dict[f"{local_idx + self.layer_idx_start}.{name}"] = v
        return new_state_dict

    def load_state_dict(
        self, state_dict, strict: bool = True, assign: bool = False
    ):
        new_state_dict = {}
        for k, v in state_dict.items():
            name = k.split(".", 1)[1]
            global_idx = int(k.split(".")[0])
            new_state_dict[
                f"layers.{global_idx - self.layer_idx_start}.{name}"
            ] = v
        return super().load_state_dict(
            new_state_dict,
            strict=strict,
            assign=assign,
        )

    def build_reparallelization_plan(
        self,
        from_model_name: ModelName,
        to_model_name: ModelName,
        from_topo: topology.PipeModelDataParallelTopology,
        to_topo: topology.PipeModelDataParallelTopology,
        to_model_config: model_api.ReaLModelConfig,
        pg_info: NCCLProcessGroupInfo,
        from_model_config: None | model_api.ReaLModelConfig = None,
    ):
        if from_model_config is None:
            from_model_config = self.config
        to_layer_mapping = partition_pipeline_layers(
            to_model_config,
            to_topo.get_dim("pipe"),
            real_model_embed_param_count,
            real_model_tblock_param_count,
            real_model_head_param_count,
        )
        to_layers_handle_dict = {}
        to_layer_indices = []
        if constants.has_model_name(to_model_name):
            with constants.model_scope(to_model_name):
                to_pp_rank = constants.pipe_parallel_rank()
                to_layer_indices = list(
                    range(
                        to_layer_mapping[to_pp_rank][0],
                        to_layer_mapping[to_pp_rank][1],
                    )
                )
                for _to_layer_idx in to_layer_indices:
                    l = self._build_layer(_to_layer_idx, to_model_config)
                    for v in l.parameters():
                        v.data = torch.tensor(
                            (), dtype=self.dtype, device=self.device
                        )
                    to_layers_handle_dict[_to_layer_idx] = l
        to_param_spec, to_param_size = build_param_spec(
            to_layer_indices,
            to_model_config,
            to_topo.get_dim("model"),
            to_topo.sequence_parallel,
        )
        if len(to_layer_indices) > 0:
            to_layer_idx_start = min(to_layer_indices)
            to_layer_idx_end = max(to_layer_indices) + 1
        else:
            to_layer_idx_start = to_layer_idx_end = -1
        to_layers_handle = nn.ModuleList(
            [to_layers_handle_dict[i] for i in to_layer_indices]
        )

        comm_plan = _derive_reparallelize_comm_plan(
            from_model_name=from_model_name,
            to_model_name=to_model_name,
            from_topo=from_topo,
            to_topo=to_topo,
            from_model_config=from_model_config,
            to_model_config=to_model_config,
            pg_info=pg_info,
            dtype=self.dtype,
        )
        rtgt = ReparallelizeTraget(
            comm_plan=comm_plan,
            to_param_spec=to_param_spec,
            to_param_size=to_param_size,
            to_layers_handle=to_layers_handle,
            to_layer_start_idx=to_layer_idx_start,
            to_layer_end_idx=to_layer_idx_end,
        )
        self._reparallelize_targets[(from_model_name, to_model_name)] = rtgt

    # FIXME: we can get topo given model name from constants
    @cuda_tmark("param_realloc", CUDATimeMarkType.mem_layout)
    def build_reparallelized_layers_async(
        self,
        from_model_name: ModelName,
        to_model_name: ModelName,
        from_topo: topology.PipeModelDataParallelTopology,
        to_topo: topology.PipeModelDataParallelTopology,
        to_model_config: model_api.ReaLModelConfig,
        pg_info: NCCLProcessGroupInfo,
    ) -> Tuple[nn.ModuleList, torch.Tensor, torch.Tensor]:
        """Trigger the parameter realloaction from the source model to the target model."""

        assert not (
            is_trainable(from_model_name) and is_trainable(to_model_name)
        )

        if (from_model_name, to_model_name) not in self._reparallelize_targets:
            self.build_reparallelization_plan(
                from_model_name,
                to_model_name,
                from_topo,
                to_topo,
                to_model_config,
                pg_info,
            )
        rtgt = self._reparallelize_targets[(from_model_name, to_model_name)]

        # Since the default implementation of PyTorch optimizers holds
        # the reference of trainable parameters, we cannot deallocate
        # them even after parameter reallocation. Therefore, there is no
        # need to release and re-allocate the trainable parameters back-and-forth.
        # We simply store the layer handles and fetch them when converting back.
        with constants.model_scope(from_model_name):
            from_model_ranks = constants.parallelism_group_ranks()
        with constants.model_scope(to_model_name):
            to_model_ranks = constants.parallelism_group_ranks()
        if (
            is_trainable(from_model_name)
            and torch.distributed.get_rank() in from_model_ranks
        ):
            store_trainable_params(
                from_model_name, (self.layers, self.contiguous_param)
            )
        elif is_trainable(to_model_name):
            del self.layers, self.contiguous_param
            self.layers = self.contiguous_param = None
            if torch.distributed.get_rank() in to_model_ranks:
                layers, param = fetch_trainable_params(to_model_name)
            else:
                layers, param = None, None
            return layers, param, 0.0

        # The following tensor holds the contiguous memory of incoming parameters
        # If this process is not a receiver, to_param_size is 0 and it's an empty tensor.
        to_contiguous_param = torch.empty(
            rtgt.to_param_size,
            dtype=self.dtype,
            device="cuda",
        )
        map_param_to_contigous_memory(
            rtgt.to_layers_handle,
            rtgt.to_param_spec,
            to_contiguous_param,
            rtgt.to_layer_start_idx,
        )

        # Allocate tensors in advance to reduce overhead.
        recv_buf_specs = []
        send_buf_specs = []
        comm_volume = torch.zeros((), dtype=torch.long, device="cuda")
        for step in rtgt.comm_plan:
            if (
                isinstance(step, ReparallelizeReceiverStep)
                and step.rank == torch.distributed.get_rank()
            ):
                if step.rank == step.src:
                    buf = slice_intervals(
                        self.contiguous_param,
                        step.sender_param_intervals,
                        step.sender_param_intervals_cpu,
                        max_interval_size=step.sender_max_interval_size,
                        output_size=step.param_size,
                    )
                else:
                    buf = torch.zeros(
                        step.param_size, dtype=step.param_dtype, device="cuda"
                    )
                    comm_volume += buf.numel()

                recv_buf_specs.append(
                    dict(
                        src=buf,
                        dst=to_contiguous_param,
                        intervals=step.receiver_param_intervals,
                        intervals_cpu=step.receiver_param_intervals_cpu,
                        max_interval_size=step.receiver_max_interval_size,
                    )
                )

            if (
                isinstance(step, ReparallelizeSenderStep)
                and step.rank == torch.distributed.get_rank()
            ):
                if step.group is not None:
                    buf = slice_intervals(
                        self.contiguous_param,
                        step.param_intervals,
                        step.param_intervals_cpu,
                        step.max_interval_size,
                        step.param_size,
                    )
                    send_buf_specs.append(buf)
                # if step.remove:
                #     for param_key in step.param_keys:
                #         layer_idx, k = param_key.split(".", 1)
                #         layer_idx = int(layer_idx)
                #         dummy_tensor = torch.tensor((), dtype=self.dtype, device=self.device)
                #         recursive_getattr(self.layers[layer_idx - self.layer_idx_start],
                #                           k).data = dummy_tensor

        # Run boradcast!
        streams = [torch.cuda.Stream() for step in rtgt.comm_plan]
        recv_buf_cnt = 0
        recv_events = []
        for step, s in zip(rtgt.comm_plan, streams):
            with torch.cuda.stream(s):
                if (
                    isinstance(step, ReparallelizeReceiverStep)
                    and step.rank == torch.distributed.get_rank()
                ):
                    e = torch.cuda.Event()
                    if step.rank != step.src:
                        buf = recv_buf_specs[recv_buf_cnt]["src"]
                        torch.distributed.broadcast(
                            buf, src=step.src, group=step.group
                        )
                    e.record(s)
                    recv_events.append(e)
                    recv_buf_cnt += 1

                if (
                    isinstance(step, ReparallelizeSenderStep)
                    and step.rank == torch.distributed.get_rank()
                ):
                    if step.group is not None:
                        buf = send_buf_specs.pop(0)
                        torch.distributed.broadcast(
                            buf, src=step.rank, group=step.group
                        )

        # Post-processing.
        assert len(send_buf_specs) == 0, len(send_buf_specs)
        assert recv_buf_cnt == len(recv_buf_specs), (
            len(recv_buf_specs),
            recv_buf_cnt,
        )
        # assert len(state_dict) == 0
        assert len(recv_events) == len(recv_buf_specs)
        for e, x in zip(recv_events, recv_buf_specs):
            torch.cuda.current_stream().wait_event(e)
            set_intervals(**x)

        # release the local GPU memory
        self.contiguous_param = None
        self.layers = None
        return rtgt.to_layers_handle, to_contiguous_param, comm_volume

    def patch_reparallelization(self, x):
        self.layers, self.contiguous_param = x
        for l in self.layers:
            for p in l.parameters():
                p.requires_grad_()


# a helper function to make real_model look like huggingface model
def generate_helper(
    self: ReaLModel,
    tokenizer: transformers.PreTrainedTokenizerFast,
    input_ids: Optional[torch.Tensor] = None,
    attention_mask: Optional[torch.Tensor] = None,
    packed_input_ids: Optional[torch.Tensor] = None,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
    gconfig: GenerationConfig = dataclasses.field(
        default_factory=GenerationConfig
    ),
) -> DuckGenerationOutput:
    assert (
        (packed_input_ids is None)
        == (cu_seqlens is None)
        == (max_seqlen is None)
    )
    if attention_mask is None and input_ids is not None:
        attention_mask = torch.ones_like(input_ids)
    if packed_input_ids is None and attention_mask is not None:
        packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
            input_ids, attention_mask
        )
    current_forward = self.forward
    self.forward = functools.partial(ReaLModel.forward, self)
    seq, scores, mask, _, _ = generate(
        model=self,
        tokenizer=tokenizer,
        packed_input_ids=packed_input_ids,
        cu_seqlens=cu_seqlens,
        max_seqlen=max_seqlen,
        gconfig=gconfig,
    )
    self.forward = current_forward
    return DuckGenerationOutput(seq, scores, mask)


# a helper function to make real_model look like huggingface model
def forward_helper(
    self: ReaLModel,
    input_ids: Optional[torch.Tensor] = None,
    attention_mask: Optional[torch.Tensor] = None,
    packed_input_ids: Optional[torch.Tensor] = None,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
) -> DuckModelOutput:
    assert (
        (packed_input_ids is None)
        == (cu_seqlens is None)
        == (max_seqlen is None)
    )
    build_packed = False
    if attention_mask is None and input_ids is not None:
        attention_mask = torch.ones_like(input_ids)
    if packed_input_ids is None and attention_mask is not None:
        build_packed = True
        packed_input_ids, indices, cu_seqlens, max_seqlen = unpad_input(
            input_ids, attention_mask
        )
        batch_size, seqlen = input_ids.shape[:2]
    x = PipeTransferData(cu_seqlens=cu_seqlens, max_seqlen=max_seqlen)
    ys = [PipeCacheData(packed_input_ids=packed_input_ids)] + [
        PipeCacheData() for _ in range(self.config.n_layers + 1)
    ]
    scores = ReaLModel.forward(self, x, ys)[0].pp_output
    if build_packed:
        scores = pad_input(scores, indices, batch_size, seqlen)
    return DuckModelOutput(logits=scores)


def add_helper_functions(m: ReaLModel):
    m.forward = functools.partial(forward_helper, m)
    m.generate = functools.partial(generate_helper, m)
    return m


def make_real_model(
    name: ModelName,
    device: torch.device,
    model_path: str,
    is_critic: bool,
    init_critic_from_actor: bool,
    dtype: Optional[str] = None,
    hf_model_family: Optional[str] = None,
) -> model_api.Model:
    if dtype == "fp16" or dtype == None:
        dtype = torch.float16
    elif dtype == "bf16":
        dtype = torch.bfloat16
    elif dtype == "fp32":
        dtype == torch.float32
    else:
        raise NotImplementedError(f"Unsupported dtype {dtype}")

    tokenizer = model_api.load_hf_tokenizer(model_path)
    mconfig = getattr(ReaLModel, f"config_from_{hf_model_family}")(
        model_path=model_path,
        is_critic=is_critic,
    )
    m = ReaLModel(mconfig, dtype=dtype, device=device)
    m._instantiation_hooks.append(
        lambda: getattr(m, f"from_{hf_model_family}")(
            load_dir=model_path, init_critic_from_actor=init_critic_from_actor
        )
    )

    if constants.pipe_parallel_world_size() == 1:
        m = add_helper_functions(m)
    return model_api.Model(name, m, tokenizer, device, dtype=dtype)


model_api.register_model("real_model", make_real_model)



from typing import Optional, Set
import dataclasses
import os
import pickle

import realhf.base.constants as constants

RECOVER_INFO_PATH = None


@dataclasses.dataclass
class StepInfo:
    epoch: int
    epoch_step: int
    global_step: int


@dataclasses.dataclass
class RecoverInfo:
    recover_start: StepInfo
    last_step_info: StepInfo
    hash_vals_to_ignore: Set[int] = dataclasses.field(default_factory=set)


def dump_recover_info(recover_info: RecoverInfo):
    global RECOVER_INFO_PATH
    if RECOVER_INFO_PATH is None:
        RECOVER_INFO_PATH = os.path.join(
            constants.RECOVER_ROOT,
            constants.experiment_name(),
            constants.trial_name(),
            "recover_info.pkl",
        )
    with open(RECOVER_INFO_PATH, "wb") as f:
        pickle.dump(recover_info, f)


def load_recover_info() -> Optional[RecoverInfo]:
    global RECOVER_INFO_PATH
    if RECOVER_INFO_PATH is None:
        RECOVER_INFO_PATH = os.path.join(
            constants.RECOVER_ROOT,
            constants.experiment_name(),
            constants.trial_name(),
            "recover_info.pkl",
        )
    try:
        with open(RECOVER_INFO_PATH, "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Resume info not found at {RECOVER_INFO_PATH}. "
            f"This should not be a resumed experiment!"
        )



from collections import defaultdict
from typing import List, Optional, Union
import os
import pickle
import time

import pandas as pd
import torch
import torch.distributed as dist
import transformers

from realhf.api.core.model_api import ReaLModelConfig
import realhf.api.core.model_api as model_api
import realhf.api.core.system_api as config_package
import realhf.base.constants as constants
import realhf.base.logging as logging

# from base.topology import *


try:
    from flash_attn.bert_padding import unpad_input
except ModuleNotFoundError:
    pass

logger = logging.getLogger("profile layers", "system")


def make_layers(config: ReaLModelConfig, dtype, device):
    from realhf.impl.model.nn.real_llm_base import (
        OutputHead,
        ReaLModelBlock,
        VocabPositionEmbedding,
    )

    embedding_layer = VocabPositionEmbedding(
        config,
        dtype=dtype,
        device=device,
    )
    real_model_blocks = [
        ReaLModelBlock(
            config,
            layer_index=i,
            output_layernorm=(i == 1),
            dtype=dtype,
            device=device,
        )
        for i in range(1)
    ]
    head = OutputHead(
        config.hidden_dim,
        1 if config.is_critic else config.vocab_size,
        bias=False,
        device=device,
        dtype=dtype,
    )

    layer_names = ["embedding_layer", "block_0", "head"]
    return [embedding_layer] + real_model_blocks + [head], layer_names


class ProfileLayers:

    def __init__(
        self,
        model_name: str,
        config: ReaLModelConfig,
        tokenizer: transformers.PreTrainedTokenizerFast = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        self.model_name = model_name
        self.config = config
        self.backend_config = config_package.ModelBackend(
            type_="deepspeed",
            args=dict(
                optimizer_name="adam",
                optimizer_config=dict(
                    lr=1e-5, weight_decay=0.0, betas=(0.9, 0.95)
                ),
                warmup_steps_proportion=0.0,
                min_lr_ratio=0.0,
                zero_stage=1,
                enable_fp16=True,
                enable_bf16=False,
            ),
        )

        self.dtype = dtype
        self.device = device
        self.layers, self.layer_names = make_layers(config, dtype, device)
        self.hidden_dim = config.hidden_dim
        self.head_dim = config.head_dim
        self.max_new_tokens = 128  # only useful in kv cache memory alloc
        self.min_new_tokens = 128

        self.stats = defaultdict(list)
        self.num_layers = len(self.layers)

        self.layers = [
            model_api.Model(name, layer, tokenizer, device=device, dtype=dtype)
            for layer, name in zip(self.layers, self.layer_names)
        ]
        self.backend = model_api.make_backend(self.backend_config)
        ft_spec = model_api.FinetuneSpec(10, 100, 10)
        self.layers = [
            self.backend.initialize(layer, ft_spec) for layer in self.layers
        ]
        self.stats = defaultdict(list)

    def reset_stats(self):
        self.stats = defaultdict(list)

    def insert_data_point(self, layer_name, name, bs, seq_len, time_ns):
        self.stats["layer_name"].append(layer_name)
        self.stats["op_name"].append(name)
        self.stats["bs"].append(bs)
        self.stats["seq_len"].append(seq_len)
        self.stats["time_ns"].append(time_ns)

    @torch.no_grad()
    def fwd_gen(self, bs, seq_len):
        from realhf.impl.model.nn.real_llm_base import (
            PipeCacheData,
            PipeTransferData,
        )

        input_ids = torch.randint(
            0,
            self.config.vocab_size,
            (bs, seq_len),
            dtype=torch.long,
            device=self.device,
        )
        attention_mask = torch.ones_like(input_ids, device=self.device)
        # fwd_gen_0
        packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
            input_ids, attention_mask
        )
        cu_seqlens = cu_seqlens.to(device=self.device)
        packed_input_ids = packed_input_ids.to(device=self.device)
        x = PipeTransferData(
            cu_seqlens=cu_seqlens,
            max_seqlen=int(max_seqlen),
            store_kv_cache=True,
        )
        ys = [PipeCacheData() for _ in range(self.num_layers)]
        ys[0].packed_input_ids = packed_input_ids

        for layer_name, layer, y in zip(self.layer_names, self.layers, ys):
            st = time.monotonic_ns()
            x: PipeTransferData = layer.module(x, y)
            x.pp_input = x.pp_output
            torch.cuda.synchronize()
            self.insert_data_point(
                layer_name, "fwd_gen_0", bs, seq_len, time.monotonic_ns() - st
            )

        prompt_logits = x.pp_output
        logits = prompt_logits[cu_seqlens[1:] - 1]
        input_lens = cu_seqlens[1:] - cu_seqlens[:-1]
        cache_seqlens = input_lens.clone().to(dtype=torch.int32)
        layer_indices = range(len(ys))

        for y, layer_idx in zip(ys[1:-1], layer_indices[1:-1]):
            assert (
                y.k_cache is not None
                and y.v_cache is not None
                and y.cache_seqlens is not None
            )
            kvcache_seqlen = max(
                max_seqlen + self.max_new_tokens,
                self.hidden_dim // self.head_dim + 10,
            )
            # fix of a flash attention bug
            k_cache = torch.zeros(
                (bs, kvcache_seqlen, *y.k_cache.shape[1:]),
                dtype=y.k_cache.dtype,
                device=self.device,
            )
            v_cache = torch.zeros_like(k_cache)
            indices = (
                torch.arange(
                    kvcache_seqlen,
                    device=torch.cuda.current_device(),
                    dtype=torch.long,
                )[None, :]
                < input_lens[:, None]
            )
            k_cache[indices] = y.k_cache
            v_cache[indices] = y.v_cache
            y.k_cache = k_cache
            y.v_cache = v_cache
            y.cache_seqlens = cache_seqlens
        x = PipeTransferData(store_kv_cache=True)
        ys[0].cache_seqlens = cache_seqlens

        # fwd_gen_1
        new_tokens = torch.randint(
            0,
            self.config.vocab_size,
            (bs,),
            dtype=torch.long,
            device=self.device,
        )
        ys[0].packed_input_ids = new_tokens
        ys[0].packed_position_ids = None
        x.cu_seqlens = torch.arange(
            bs + 1, dtype=torch.int32, device=self.device
        )
        x.max_seqlen = 1
        for layer_name, layer, y in zip(self.layer_names, self.layers, ys):
            st = time.monotonic_ns()
            x = layer.module(x, y)
            x.pp_input = x.pp_output
            torch.cuda.synchronize()
            self.insert_data_point(
                layer_name, "fwd_gen_1", bs, seq_len, time.monotonic_ns() - st
            )

    def fwd_bwd_opt(self, bs, seq_len):
        from realhf.impl.model.nn.real_llm_base import (
            PipeCacheData,
            PipeTransferData,
        )

        input_ids = torch.randint(
            0,
            self.config.vocab_size,
            (bs, seq_len),
            dtype=torch.long,
            device=self.device,
        )
        attention_mask = torch.ones_like(input_ids, device=self.device)
        packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
            input_ids, attention_mask
        )
        cu_seqlens = cu_seqlens.to(device=self.device)
        packed_input_ids = packed_input_ids.to(device=self.device)
        x = PipeTransferData(
            cu_seqlens=cu_seqlens, max_seqlen=max_seqlen, store_kv_cache=False
        )
        ys = [PipeCacheData() for _ in range(self.num_layers)]
        ys[0].packed_input_ids = packed_input_ids

        for layer_name, layer, y in zip(self.layer_names, self.layers, ys):
            # fwd
            st = time.monotonic_ns()
            x: PipeTransferData = layer.module(x, y)
            torch.cuda.synchronize()
            self.insert_data_point(
                layer_name, "fwd", bs, seq_len, time.monotonic_ns() - st
            )
            # bwd
            r = torch.rand(
                *x.pp_output.shape,
                device=x.pp_output.device,
                dtype=x.pp_output.dtype,
            )
            loss = torch.max(x.pp_output * r)
            st = time.monotonic_ns()
            layer.module.backward(loss)
            torch.cuda.synchronize()
            self.insert_data_point(
                layer_name, "bwd", bs, seq_len, time.monotonic_ns() - st
            )
            # opt
            st = time.monotonic_ns()
            layer.module.step()
            torch.cuda.synchronize()
            self.insert_data_point(
                layer_name, "opt", bs, seq_len, time.monotonic_ns() - st
            )
            x.pp_input = x.pp_output.clone().detach()

    def make_dataframe_and_print(self):
        df = pd.DataFrame(self.stats)
        logger.info(f"Current Stats: \nstr{df}")

    def dump_stats(self, world_size):
        rank = dist.get_rank()
        # dump full stats
        dump_dir = os.path.join(
            constants.PROFILER_CACHE_PATH,
            "layer_stats",
        )
        dump_path = os.path.join(
            dump_dir, self.model_name, f"layer-stats_{world_size}_{rank}.pkl"
        )
        os.makedirs(os.path.dirname(dump_path), exist_ok=True)

        with open(dump_path, "wb") as f:
            df = pd.DataFrame(self.stats)
            pickle.dump(df, f)


def make_profile_layers(
    device: torch.device,
    model_path: str,
    model_name: str,
    dtype: Optional[str] = None,
    hf_model_type: str = "llama",
):
    from realhf.impl.model.nn.real_llm_api import ReaLModel

    if dtype == "fp16" or dtype == None:
        dtype = torch.float16
    elif dtype == "bf16":
        dtype = torch.bfloat16
    elif dtype == "fp32":
        dtype == torch.float32
    else:
        raise NotImplementedError(f"Unsupported dtype {dtype}")
    tokenizer = None
    config: ReaLModelConfig = getattr(
        ReaLModel, f"config_from_{hf_model_type}"
    )(
        model_path=model_path,
    )
    if tokenizer is None:
        tokenizer = model_api.load_hf_tokenizer(model_path)

    profile_layers = ProfileLayers(
        model_name, config, tokenizer=tokenizer, dtype=dtype, device=device
    )

    return profile_layers



from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union
import os
import re
import signal as signal_module
import subprocess
import time

import psutil

from realhf.base.constants import LOG_ROOT
from realhf.scheduler.client import (
    JobException,
    JobInfo,
    JobState,
    SchedulerClient,
    SchedulerError,
)
import realhf.base.logging as logging

logger = logging.getLogger("Local Scheduler")

JOB_STATE_TO_PROCESS_STATUS = {
    JobState.NOT_FOUND: [],
    JobState.PENDING: [psutil.STATUS_PARKED],
    JobState.RUNNING: [
        psutil.STATUS_RUNNING,
        psutil.STATUS_SLEEPING,
        psutil.STATUS_DISK_SLEEP,
        psutil.STATUS_TRACING_STOP,
        psutil.STATUS_WAKING,
        psutil.STATUS_WAITING,
        psutil.STATUS_LOCKED,
        psutil.STATUS_IDLE,
    ],
    JobState.COMPLETED: [
        psutil.STATUS_DEAD,
        psutil.STATUS_STOPPED,
        psutil.STATUS_ZOMBIE,
    ],
    JobState.FAILED: [],
    JobState.CANCELLED: [],
}

PROCESS_STATUS_TO_JOB_STATE = {}
for job_state, process_statuses in JOB_STATE_TO_PROCESS_STATUS.items():
    for process_status in process_statuses:
        PROCESS_STATUS_TO_JOB_STATE[process_status] = job_state


def terminate_process_and_children(
    pid: int, signal: Optional[Union[str, int]] = None
):
    if signal is None:
        signal = signal_module.SIGKILL
    if isinstance(signal, str):
        signal = getattr(signal_module, signal)
    try:
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        for child in children:
            terminate_process_and_children(child.pid)
        parent.send_signal(signal)
    except psutil.NoSuchProcess:
        pass


class LocalSchedulerClient(SchedulerClient):
    """Instead of talking to the scheduler server (the typical behaviour), this client starts jobs directly
    on the local host and keeps a collection of job processes.
    """

    def log_path_of(self, worker_type) -> str:
        return os.path.join(
            LOG_ROOT,
            self.expr_name,
            self.trial_name,
            f"{worker_type}-0",
        )

    def __init__(self, expr_name, trial_name):
        super().__init__(expr_name, trial_name)
        self._jobs: Dict[str, subprocess.Popen] = {}
        self._running_worker_types = []

        self._gpu_counter = 0
        self._cuda_devices: List[str] = os.environ.get(
            "CUDA_VISIBLE_DEVICES", ""
        ).split(",")

        self._job_counter: Dict[str, int] = defaultdict(int)
        self._job_with_gpu: Dict[str, bool] = defaultdict(int)
        self._job_env_vars: Dict[str, Dict] = defaultdict(int)
        self._job_cmd = {}
        self._job_states = {}

        if len(self._cuda_devices) < 1:
            raise RuntimeError(
                f"Local mode can only run when there is at least one GPU. "
                f"CUDA_VISIBLE_DEVICES is currently set to {os.environ['CUDA_VISIBLE_DEVICES']}."
            )

    def __del__(self):
        self.wait(commit=False)

    def submit_array(
        self,
        worker_type: str,
        cmd: str,
        count: int = 1,
        gpu: int = 0,
        env_vars: Optional[Dict] = None,
        **kwargs,
    ):
        if env_vars is None:
            env_vars = {}

        self._job_counter[worker_type] += count
        if worker_type in self._job_with_gpu:
            assert self._job_with_gpu[worker_type] == (
                gpu > 0
            ), "All workers of the same type must either use GPU or not use GPU."
        else:
            self._job_with_gpu[worker_type] = gpu > 0

        if worker_type in self._job_env_vars:
            assert (
                self._job_env_vars[worker_type] == env_vars
            ), "All workers of the same type must have the same env vars."
        else:
            self._job_env_vars[worker_type] = env_vars

        if worker_type in self._job_cmd:
            assert (
                self._job_cmd[worker_type] == cmd
            ), "All workers of the same type must have the same cmd."
        else:
            self._job_cmd[worker_type] = cmd

    def submit(self, worker_type, cmd, **kwargs):
        self.submit_array(worker_type, cmd, count=1, **kwargs)

    def __commit_all(self):
        for worker_type, count, use_gpu, env_vars in zip(
            self._job_counter.keys(),
            self._job_counter.values(),
            self._job_with_gpu.values(),
            self._job_env_vars.values(),
        ):
            os.makedirs(
                os.path.dirname(self.log_path_of(worker_type)),
                exist_ok=True,
                mode=0o775,
            )
            for i in range(count):
                if use_gpu:
                    available_device_id = self._gpu_counter % len(
                        self._cuda_devices
                    )
                    env_vars["CUDA_VISIBLE_DEVICES"] = str(
                        self._cuda_devices[available_device_id]
                    )
                    self._gpu_counter += 1
                cmd = (
                    " ".join(str(k) + "=" + str(v) for k, v in env_vars.items())
                    + " stdbuf -oL "
                    + self._job_cmd[worker_type]
                )
                # Run `apps.remote` with a single process.
                # This simulates a multi-prog slurm job with `count` jobsteps, with each jobstep having a single process.
                cmd = cmd.format(
                    jobstep_id=i,
                    n_jobsteps=count,
                    worker_submission_index=0,
                    wprocs_per_jobstep=1,
                    wprocs_in_job=count,
                    wproc_offset=0,
                )
                logger.debug("Starting local process with command: %s", cmd)
                cmd = f"{cmd} | tee -a {self.log_path_of(worker_type)}"
                process = subprocess.Popen(cmd, shell=isinstance(cmd, str))
                self._jobs[f"{worker_type}/{i}"] = process
            self._running_worker_types.append(worker_type)

    def stop(self, worker_type, signal=None):
        assert any(k.startswith(worker_type) for k in self._jobs)
        keys = [k for k, p in self._jobs.items() if k.startswith(worker_type)]
        procs = [p for k, p in self._jobs.items() if k.startswith(worker_type)]
        logger.info(
            f"Stopping local process with signal {signal if signal else 'SIGKILL'}, "
            f"pid: {[p.pid for p in procs]}"
        )
        for p in procs:
            terminate_process_and_children(p.pid, signal=signal)
        for p in procs:
            p.wait()
        for k, p in zip(keys, procs):
            self._jobs.pop(k)
            del p
        self._running_worker_types.remove(worker_type)

    def stop_all(self, signal=None):
        # signal argument is ignored in local stop_all
        for name in self._job_counter:
            self.stop(name, signal=signal)

    def find(self, job_name):
        if job_name in self._jobs:
            return JobInfo(
                name=job_name, state=JobState.RUNNING, host="localhost"
            )
        else:
            return JobInfo(name=job_name, state=JobState.NOT_FOUND)

    def find_all(self, job_name_regex=".*"):
        rs = []
        for name in self._jobs:
            if re.fullmatch(job_name_regex, name):
                rs.append(self.find(name))
        return rs

    def wait(
        self,
        timeout=None,
        check_status: Tuple[JobState, ...] = (
            JobState.CANCELLED,
            JobState.FAILED,
            JobState.NOT_FOUND,
        ),
        remove_status: Tuple[JobState, ...] = (JobState.COMPLETED,),
        update=False,
        commit=True,
    ):
        if commit:
            self.__commit_all()
        deadline = None if timeout is None else time.time() + timeout
        logger.info(
            "Waiting for %d local running processes, pids: %s",
            len(self._jobs),
            " ".join(str(job.pid) for job in self._jobs.values()),
        )
        left = set(self._jobs.keys())
        num_jobs_left = len(left)

        while len(left) > 0:
            to_remove = []
            if len(left) < num_jobs_left:
                num_jobs_left = len(left)
                logger.info(f"Waiting for {num_jobs_left} jobs.")
            if deadline is not None and time.time() > deadline:
                raise TimeoutError(
                    f"Timeout waiting for {self.run_name}: {', '.join(sorted(left))}"
                )
            # update job states
            for job_name in list(left):
                job = self._jobs[job_name]
                pid = job.pid
                process = psutil.Process(pid)
                self._job_states[job_name] = PROCESS_STATUS_TO_JOB_STATE.get(
                    process.status(), JobState.NOT_FOUND
                )

            for job_name in list(left):
                state = self._job_states[job_name]
                if state in check_status:
                    raise JobException(
                        run_name=self.run_name,
                        worker_type=job_name.split("/")[0],
                        host="local",
                        reason=state,
                    )
                if state in remove_status:
                    logger.info(f"Job {job_name} is {state}.(Removed)")
                    left.remove(job_name)
                    to_remove.append(job_name)

            if update:
                for k in to_remove:
                    self._jobs.pop(k)
                    worker_type = k.split("/")[0]
                    assert worker_type in self._job_counter
                    self._job_counter[worker_type] -= 1
                    if self._job_counter[worker_type] <= 0:
                        assert worker_type in self._job_with_gpu
                        assert worker_type in self._job_env_vars
                        assert worker_type in self._job_cmd
                        self._job_counter.pop(worker_type)
                        self._job_with_gpu.pop(worker_type)
                        self._job_env_vars.pop(worker_type)
                        self._job_cmd.pop(worker_type)

            time.sleep(2)
